"""
RLM-Graph: Knowledge graph-based context storage and retrieval.

Extends the RLM system with graph-structured storage where:
- Nodes represent documents, sections, chunks, and entities
- Edges capture relationships (has_section, has_chunk, mentions, related_to)
- Graph traversal enables finding structurally related context
"""

from __future__ import annotations

import hashlib
import logging
import re
import uuid
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Any, Literal

from pinkyclawd.config.storage import ContextBlock, Message, PartType, get_storage

logger = logging.getLogger(__name__)


class GraphStorageAdapter:
    """
    Adapter that provides a unified interface for graph storage operations.

    Supports both JSON-based and SQLite storage backends with the same interface.
    Uses the method names expected by GraphIngester/Traverser/Searcher.
    """

    def __init__(self, session_id: str | None = None) -> None:
        """
        Initialize the storage adapter.

        Args:
            session_id: Session ID for JSON storage (required for JSON backend)
        """
        self._session_id = session_id
        self._storage_type: str | None = None
        self._json_storage: Any = None
        self._sqlite_storage: Any = None

    def _get_storage_type(self) -> str:
        """Get the configured storage type."""
        if self._storage_type is None:
            try:
                from pinkyclawd.config.settings import get_config
                config = get_config()
                self._storage_type = config.rlm.graph_storage
            except Exception:
                self._storage_type = "json"  # Default to JSON
        return self._storage_type

    def _get_json_storage(self, session_id: str):
        """Get or create JSON storage for a session."""
        from pinkyclawd.rlm.graph_json import get_json_graph_storage
        return get_json_graph_storage(session_id)

    def _get_sqlite_storage(self):
        """Get the SQLite storage instance."""
        if self._sqlite_storage is None:
            self._sqlite_storage = get_storage()
        return self._sqlite_storage

    def add_graph_node(
        self,
        node_id: str,
        session_id: str,
        node_type: str,
        content: str,
        metadata: dict[str, Any] | None = None,
        embedding: list[float] | None = None,
    ) -> None:
        """Add a node to the knowledge graph."""
        if self._get_storage_type() == "json":
            storage = self._get_json_storage(session_id)
            storage.add_node(node_id, session_id, node_type, content, metadata, embedding)
        else:
            self._get_sqlite_storage().add_graph_node(
                node_id, session_id, node_type, content, metadata, embedding
            )

    def get_graph_node(self, node_id: str) -> dict[str, Any] | None:
        """Get a node from the knowledge graph."""
        if self._get_storage_type() == "json":
            # Need to search across all sessions for JSON storage
            # This is less efficient but maintains compatibility
            if self._session_id:
                storage = self._get_json_storage(self._session_id)
                return storage.get_node(node_id)
            return None
        else:
            return self._get_sqlite_storage().get_graph_node(node_id)

    def get_graph_nodes(
        self,
        session_id: str | None = None,
        node_type: str | None = None,
        limit: int = 100,
    ) -> list[dict[str, Any]]:
        """Get nodes from the knowledge graph with optional filters."""
        if self._get_storage_type() == "json":
            if session_id:
                storage = self._get_json_storage(session_id)
                return storage.get_nodes(session_id, node_type, limit)
            elif self._session_id:
                storage = self._get_json_storage(self._session_id)
                return storage.get_nodes(self._session_id, node_type, limit)
            return []
        else:
            return self._get_sqlite_storage().get_graph_nodes(session_id, node_type, limit)

    def search_graph_nodes(
        self,
        query: str,
        session_id: str | None = None,
        node_type: str | None = None,
        limit: int = 20,
    ) -> list[dict[str, Any]]:
        """Search nodes by content."""
        if self._get_storage_type() == "json":
            sid = session_id or self._session_id
            if sid:
                storage = self._get_json_storage(sid)
                return storage.search_nodes(query, session_id, node_type, limit)
            return []
        else:
            return self._get_sqlite_storage().search_graph_nodes(
                query, session_id, node_type, limit
            )

    def delete_graph_node(self, node_id: str) -> bool:
        """Delete a node and its edges."""
        if self._get_storage_type() == "json":
            if self._session_id:
                storage = self._get_json_storage(self._session_id)
                return storage.delete_node(node_id)
            return False
        else:
            return self._get_sqlite_storage().delete_graph_node(node_id)

    def add_graph_edge(
        self,
        source_id: str,
        target_id: str,
        relationship: str,
        weight: float = 1.0,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Add an edge to the knowledge graph."""
        if self._get_storage_type() == "json":
            if self._session_id:
                storage = self._get_json_storage(self._session_id)
                storage.add_edge(source_id, target_id, relationship, weight, metadata)
        else:
            self._get_sqlite_storage().add_graph_edge(
                source_id, target_id, relationship, weight, metadata
            )

    def get_graph_edges(
        self,
        node_id: str,
        direction: Literal["outgoing", "incoming", "both"] = "both",
        relationship: str | None = None,
    ) -> list[dict[str, Any]]:
        """Get edges connected to a node."""
        if self._get_storage_type() == "json":
            if self._session_id:
                storage = self._get_json_storage(self._session_id)
                return storage.get_edges(node_id, direction, relationship)
            return []
        else:
            return self._get_sqlite_storage().get_graph_edges(node_id, direction, relationship)

    def get_neighbors(
        self,
        node_id: str,
        relationship: str | None = None,
        direction: Literal["outgoing", "incoming", "both"] = "outgoing",
        node_type: str | None = None,
    ) -> list[dict[str, Any]]:
        """Get neighboring nodes connected by edges."""
        if self._get_storage_type() == "json":
            if self._session_id:
                storage = self._get_json_storage(self._session_id)
                return storage.get_neighbors(node_id, relationship, direction, node_type)
            return []
        else:
            return self._get_sqlite_storage().get_neighbors(
                node_id, relationship, direction, node_type
            )

    def delete_graph_edge(
        self,
        source_id: str,
        target_id: str,
        relationship: str,
    ) -> bool:
        """Delete an edge from the knowledge graph."""
        if self._get_storage_type() == "json":
            if self._session_id:
                storage = self._get_json_storage(self._session_id)
                return storage.delete_edge(source_id, target_id, relationship)
            return False
        else:
            return self._get_sqlite_storage().delete_graph_edge(
                source_id, target_id, relationship
            )

    def clear_session_graph(self, session_id: str) -> None:
        """Clear all graph data for a session."""
        if self._get_storage_type() == "json":
            storage = self._get_json_storage(session_id)
            storage.clear_session()
        else:
            self._get_sqlite_storage().clear_session_graph(session_id)


# Global storage adapter instance cache
_storage_adapters: dict[str, GraphStorageAdapter] = {}


def get_graph_storage(session_id: str | None = None) -> GraphStorageAdapter:
    """
    Get a graph storage adapter for the given session.

    Args:
        session_id: Session ID (required for JSON storage)

    Returns:
        GraphStorageAdapter instance
    """
    global _storage_adapters

    key = session_id or "_default"
    if key not in _storage_adapters:
        _storage_adapters[key] = GraphStorageAdapter(session_id)
    return _storage_adapters[key]


class NodeType(str, Enum):
    """Types of nodes in the knowledge graph."""

    DOCUMENT = "document"  # Top-level document/conversation
    SECTION = "section"  # Major section or topic
    CHUNK = "chunk"  # Smaller retrievable unit
    ENTITY = "entity"  # Extracted entity (person, concept, code element)


class RelationType(str, Enum):
    """Types of relationships between nodes."""

    HAS_SECTION = "has_section"  # Document -> Section
    HAS_CHUNK = "has_chunk"  # Section -> Chunk, or Document -> Chunk
    MENTIONS = "mentions"  # Chunk -> Entity
    RELATED_TO = "related_to"  # Entity -> Entity, or Chunk -> Chunk
    FOLLOWS = "follows"  # Temporal ordering between chunks
    PARENT_OF = "parent_of"  # Hierarchical relationship


@dataclass
class GraphNode:
    """A node in the knowledge graph."""

    id: str
    session_id: str
    type: NodeType
    content: str
    metadata: dict[str, Any] = field(default_factory=dict)
    embedding: list[float] | None = None
    created_at: datetime = field(default_factory=datetime.now)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> GraphNode:
        return cls(
            id=data["id"],
            session_id=data["session_id"],
            type=NodeType(data["type"]),
            content=data["content"],
            metadata=data.get("metadata", {}),
            embedding=data.get("embedding"),
            created_at=datetime.fromisoformat(data["created_at"])
            if isinstance(data["created_at"], str)
            else data["created_at"],
        )

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "session_id": self.session_id,
            "type": self.type.value,
            "content": self.content,
            "metadata": self.metadata,
            "embedding": self.embedding,
            "created_at": self.created_at.isoformat(),
        }


@dataclass
class GraphEdge:
    """An edge connecting two nodes."""

    source_id: str
    target_id: str
    relationship: RelationType
    weight: float = 1.0
    metadata: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        return {
            "source_id": self.source_id,
            "target_id": self.target_id,
            "relationship": self.relationship.value,
            "weight": self.weight,
            "metadata": self.metadata,
        }


class EntityExtractor:
    """
    Extracts entities from text content.

    Identifies:
    - Capitalized terms (potential proper nouns)
    - Code elements (function names, classes, variables)
    - Markdown links and references
    - Frequent keywords
    """

    # Common words to ignore
    STOPWORDS = {
        "the", "a", "an", "is", "are", "was", "were", "be", "been", "being",
        "have", "has", "had", "do", "does", "did", "will", "would", "could",
        "should", "may", "might", "must", "shall", "can", "to", "of", "in",
        "for", "on", "with", "at", "by", "from", "as", "into", "through",
        "during", "before", "after", "above", "below", "between", "this",
        "that", "these", "those", "then", "than", "when", "where", "which",
        "who", "what", "how", "why", "all", "any", "both", "each", "few",
        "more", "most", "other", "some", "such", "no", "not", "only", "same",
        "so", "than", "too", "very", "just", "also", "now", "here", "there",
        "user", "assistant", "tool", "result", "error", "success",
    }

    # Code-related patterns
    CODE_PATTERNS = [
        r"\b([A-Z][a-zA-Z0-9]*(?:Error|Exception|Handler|Manager|Service|Controller|Factory|Builder))\b",  # Classes
        r"\b(def\s+)?([a-z_][a-z0-9_]*)\s*\(",  # Functions
        r"\b([A-Z_][A-Z0-9_]{2,})\b",  # Constants
        r"`([^`]+)`",  # Inline code
        r"(?:class|interface|type|struct)\s+([A-Z][a-zA-Z0-9]*)",  # Type definitions
    ]

    def __init__(self) -> None:
        self._compiled_patterns = [re.compile(p) for p in self.CODE_PATTERNS]

    def extract(self, text: str) -> list[tuple[str, str]]:
        """
        Extract entities from text.

        Returns:
            List of (entity_name, entity_type) tuples
        """
        entities: dict[str, str] = {}

        # Extract capitalized terms (potential proper nouns)
        capitalized = re.findall(r"\b([A-Z][a-z]+(?:\s+[A-Z][a-z]+)*)\b", text)
        for term in capitalized:
            if term.lower() not in self.STOPWORDS and len(term) > 2:
                entities[term] = "proper_noun"

        # Extract code elements
        for pattern in self._compiled_patterns:
            matches = pattern.findall(text)
            for match in matches:
                if isinstance(match, tuple):
                    # Take the last non-empty group
                    term = next((m for m in reversed(match) if m), None)
                else:
                    term = match

                if term and term.lower() not in self.STOPWORDS and len(term) > 1:
                    entities[term] = "code_element"

        # Extract markdown links
        links = re.findall(r"\[([^\]]+)\]\([^)]+\)", text)
        for link_text in links:
            if link_text.lower() not in self.STOPWORDS:
                entities[link_text] = "reference"

        # Extract file paths
        paths = re.findall(r"(?:^|\s)([./]?(?:[a-zA-Z0-9_-]+/)+[a-zA-Z0-9_.-]+)", text)
        for path in paths:
            entities[path] = "file_path"

        return list(entities.items())

    def extract_keywords(self, text: str, top_n: int = 10) -> list[str]:
        """Extract top keywords based on frequency."""
        words = re.findall(r"\b[a-zA-Z]{3,}\b", text.lower())
        word_counts: dict[str, int] = {}

        for word in words:
            if word not in self.STOPWORDS:
                word_counts[word] = word_counts.get(word, 0) + 1

        sorted_words = sorted(word_counts.items(), key=lambda x: x[1], reverse=True)
        return [word for word, _ in sorted_words[:top_n]]


class ContentChunker:
    """
    Splits content into hierarchical chunks.

    Creates sections based on natural breaks and chunks
    suitable for retrieval.
    """

    def __init__(
        self,
        section_min_chars: int = 500,
        chunk_target_chars: int = 300,
        chunk_overlap: int = 50,
    ) -> None:
        self.section_min_chars = section_min_chars
        self.chunk_target_chars = chunk_target_chars
        self.chunk_overlap = chunk_overlap

    def split_into_sections(self, content: str) -> list[tuple[str, str]]:
        """
        Split content into sections.

        Returns:
            List of (section_title, section_content) tuples
        """
        sections: list[tuple[str, str]] = []

        # Try to split by markdown headers
        header_pattern = r"(?:^|\n)(#{1,3})\s+(.+?)(?:\n|$)"
        matches = list(re.finditer(header_pattern, content))

        if matches:
            for i, match in enumerate(matches):
                title = match.group(2).strip()
                start = match.end()
                end = matches[i + 1].start() if i + 1 < len(matches) else len(content)
                section_content = content[start:end].strip()

                if section_content:
                    sections.append((title, section_content))
        else:
            # Fall back to paragraph-based splitting
            paragraphs = content.split("\n\n")
            current_section = []
            current_length = 0

            for para in paragraphs:
                current_section.append(para)
                current_length += len(para)

                if current_length >= self.section_min_chars:
                    section_text = "\n\n".join(current_section)
                    # Generate title from first line
                    first_line = section_text.split("\n")[0][:50]
                    sections.append((first_line, section_text))
                    current_section = []
                    current_length = 0

            if current_section:
                section_text = "\n\n".join(current_section)
                first_line = section_text.split("\n")[0][:50]
                sections.append((first_line, section_text))

        return sections

    def split_into_chunks(self, content: str) -> list[str]:
        """
        Split content into overlapping chunks.

        Returns:
            List of chunk texts
        """
        if len(content) <= self.chunk_target_chars:
            return [content]

        chunks: list[str] = []
        sentences = re.split(r"(?<=[.!?])\s+", content)
        current_chunk: list[str] = []
        current_length = 0

        for sentence in sentences:
            sentence_len = len(sentence)

            if current_length + sentence_len > self.chunk_target_chars and current_chunk:
                chunks.append(" ".join(current_chunk))

                # Keep overlap
                overlap_chunk: list[str] = []
                overlap_len = 0
                for s in reversed(current_chunk):
                    if overlap_len + len(s) <= self.chunk_overlap:
                        overlap_chunk.insert(0, s)
                        overlap_len += len(s)
                    else:
                        break

                current_chunk = overlap_chunk
                current_length = overlap_len

            current_chunk.append(sentence)
            current_length += sentence_len

        if current_chunk:
            chunks.append(" ".join(current_chunk))

        return chunks


class GraphIngester:
    """
    Ingests content into the knowledge graph.

    Converts documents into a hierarchical graph structure:
    Document -> Sections -> Chunks -> Entities
    """

    def __init__(self, session_id: str | None = None) -> None:
        self._session_id = session_id
        self._storage = get_graph_storage(session_id)
        self._chunker = ContentChunker()
        self._extractor = EntityExtractor()
        self._embedding_manager = None

    def set_session_id(self, session_id: str) -> None:
        """Set the session ID and update storage."""
        self._session_id = session_id
        self._storage = get_graph_storage(session_id)

    def _get_embedding_manager(self):
        """Lazy load embedding manager."""
        if self._embedding_manager is None:
            try:
                from pinkyclawd.rlm.embedding import get_embedding_manager
                self._embedding_manager = get_embedding_manager()
            except ImportError:
                logger.warning("Embedding module not available")
        return self._embedding_manager

    def _generate_node_id(self, node_type: str, content: str) -> str:
        """Generate a unique node ID based on content hash."""
        content_hash = hashlib.md5(content.encode()).hexdigest()[:8]
        return f"{node_type}_{uuid.uuid4().hex[:8]}_{content_hash}"

    def ingest_context_block(
        self,
        block: ContextBlock,
        generate_embeddings: bool = True,
    ) -> GraphNode:
        """
        Ingest a context block into the graph.

        Creates a document node and recursively creates
        sections, chunks, and entities.

        Returns:
            The root document node
        """
        # Ensure storage is set for this session
        if self._session_id != block.session_id:
            self.set_session_id(block.session_id)

        # Create document node
        doc_node = GraphNode(
            id=self._generate_node_id("doc", block.id),
            session_id=block.session_id,
            type=NodeType.DOCUMENT,
            content=block.summary or block.content[:200],
            metadata={
                "block_id": block.id,
                "task_id": block.task_id,
                "task_description": block.task_description,
                "tokens": block.tokens,
            },
        )

        self._storage.add_graph_node(
            node_id=doc_node.id,
            session_id=doc_node.session_id,
            node_type=doc_node.type.value,
            content=doc_node.content,
            metadata=doc_node.metadata,
        )

        # Split into sections
        sections = self._chunker.split_into_sections(block.content)

        for section_title, section_content in sections:
            section_node = self._ingest_section(
                doc_node.id,
                block.session_id,
                section_title,
                section_content,
                generate_embeddings,
            )

            # Link document -> section
            self._storage.add_graph_edge(
                source_id=doc_node.id,
                target_id=section_node.id,
                relationship=RelationType.HAS_SECTION.value,
            )

        logger.info(f"Ingested context block {block.id} as graph node {doc_node.id}")
        return doc_node

    def _ingest_section(
        self,
        parent_id: str,
        session_id: str,
        title: str,
        content: str,
        generate_embeddings: bool,
    ) -> GraphNode:
        """Ingest a section and its chunks."""
        section_node = GraphNode(
            id=self._generate_node_id("sec", content),
            session_id=session_id,
            type=NodeType.SECTION,
            content=title,
            metadata={"full_content": content[:500]},
        )

        self._storage.add_graph_node(
            node_id=section_node.id,
            session_id=session_id,
            node_type=section_node.type.value,
            content=section_node.content,
            metadata=section_node.metadata,
        )

        # Split into chunks
        chunks = self._chunker.split_into_chunks(content)
        prev_chunk_id: str | None = None

        for chunk_content in chunks:
            chunk_node = self._ingest_chunk(
                section_node.id,
                session_id,
                chunk_content,
                generate_embeddings,
            )

            # Link section -> chunk
            self._storage.add_graph_edge(
                source_id=section_node.id,
                target_id=chunk_node.id,
                relationship=RelationType.HAS_CHUNK.value,
            )

            # Link previous chunk -> current chunk (temporal order)
            if prev_chunk_id:
                self._storage.add_graph_edge(
                    source_id=prev_chunk_id,
                    target_id=chunk_node.id,
                    relationship=RelationType.FOLLOWS.value,
                )

            prev_chunk_id = chunk_node.id

        return section_node

    def _ingest_chunk(
        self,
        parent_id: str,
        session_id: str,
        content: str,
        generate_embeddings: bool,
    ) -> GraphNode:
        """Ingest a chunk and extract entities."""
        # Generate embedding if enabled
        embedding = None
        if generate_embeddings:
            embedding_manager = self._get_embedding_manager()
            if embedding_manager:
                try:
                    import asyncio
                    loop = asyncio.get_event_loop()
                    if loop.is_running():
                        # Schedule for later
                        pass
                    else:
                        embedding = loop.run_until_complete(
                            embedding_manager.embed(content)
                        )
                except Exception as e:
                    logger.debug(f"Failed to generate embedding: {e}")

        chunk_node = GraphNode(
            id=self._generate_node_id("chunk", content),
            session_id=session_id,
            type=NodeType.CHUNK,
            content=content,
            embedding=embedding,
        )

        self._storage.add_graph_node(
            node_id=chunk_node.id,
            session_id=session_id,
            node_type=chunk_node.type.value,
            content=chunk_node.content,
            metadata=chunk_node.metadata,
            embedding=embedding,
        )

        # Extract and link entities
        entities = self._extractor.extract(content)
        entity_cache: dict[str, str] = {}  # entity_name -> entity_id

        for entity_name, entity_type in entities:
            # Check if entity already exists
            if entity_name in entity_cache:
                entity_id = entity_cache[entity_name]
            else:
                # Create or reuse entity node
                entity_id = self._get_or_create_entity(
                    session_id, entity_name, entity_type
                )
                entity_cache[entity_name] = entity_id

            # Link chunk -> entity
            self._storage.add_graph_edge(
                source_id=chunk_node.id,
                target_id=entity_id,
                relationship=RelationType.MENTIONS.value,
            )

        return chunk_node

    def _get_or_create_entity(
        self,
        session_id: str,
        name: str,
        entity_type: str,
    ) -> str:
        """Get existing entity or create new one."""
        # Search for existing entity with same name in session
        existing = self._storage.search_graph_nodes(
            query=name,
            session_id=session_id,
            node_type=NodeType.ENTITY.value,
            limit=1,
        )

        if existing and existing[0]["content"] == name:
            return existing[0]["id"]

        # Create new entity
        entity_id = self._generate_node_id("ent", name)
        self._storage.add_graph_node(
            node_id=entity_id,
            session_id=session_id,
            node_type=NodeType.ENTITY.value,
            content=name,
            metadata={"entity_type": entity_type},
        )

        return entity_id

    def ingest_messages(
        self,
        session_id: str,
        messages: list[Message],
        generate_embeddings: bool = True,
    ) -> GraphNode | None:
        """
        Ingest messages directly into the graph.

        Creates a document node from the messages.
        """
        if not messages:
            return None

        # Ensure storage is set for this session
        if self._session_id != session_id:
            self.set_session_id(session_id)

        # Build content from messages
        content_parts = []
        for msg in messages:
            role = msg.role.value.upper()
            text_parts = []

            for part in msg.parts:
                if part.type == PartType.TEXT:
                    text_parts.append(part.content.get("text", ""))
                elif part.type == PartType.TOOL_USE:
                    tool_name = part.content.get("name", "unknown")
                    text_parts.append(f"[Tool: {tool_name}]")
                elif part.type == PartType.TOOL_RESULT:
                    result = part.content.get("result", "")
                    if len(result) > 300:
                        result = result[:300] + "..."
                    text_parts.append(f"[Result: {result}]")

            if text_parts:
                content_parts.append(f"{role}: {' '.join(text_parts)}")

        content = "\n\n".join(content_parts)

        # Create a temporary context block
        block = ContextBlock(
            id=f"msg_{uuid.uuid4().hex[:12]}",
            session_id=session_id,
            summary=content[:200],
            content=content,
            tokens=len(content.split()),  # Rough estimate
        )

        return self.ingest_context_block(block, generate_embeddings)


class GraphTraverser:
    """
    Traverses the knowledge graph to find related context.

    Implements various traversal strategies:
    - Breadth-first exploration
    - Semantic similarity expansion
    - Entity-based linking
    """

    def __init__(self, session_id: str | None = None) -> None:
        self._session_id = session_id
        self._storage = get_graph_storage(session_id)

    def set_session_id(self, session_id: str) -> None:
        """Set the session ID and update storage."""
        self._session_id = session_id
        self._storage = get_graph_storage(session_id)

    def get_neighbors(
        self,
        node_id: str,
        relationship: RelationType | str | None = None,
        direction: Literal["outgoing", "incoming", "both"] = "outgoing",
        node_type: NodeType | str | None = None,
    ) -> list[GraphNode]:
        """Get neighboring nodes."""
        rel_str = relationship.value if isinstance(relationship, RelationType) else relationship
        type_str = node_type.value if isinstance(node_type, NodeType) else node_type

        neighbors = self._storage.get_neighbors(
            node_id=node_id,
            relationship=rel_str,
            direction=direction,
            node_type=type_str,
        )

        return [GraphNode.from_dict(n) for n in neighbors]

    def expand_context(
        self,
        node_ids: list[str],
        max_depth: int = 2,
        max_nodes: int = 20,
    ) -> list[GraphNode]:
        """
        Expand context by traversing from given nodes.

        Uses breadth-first search to find related nodes.
        """
        visited: set[str] = set()
        result: list[GraphNode] = []
        queue: list[tuple[str, int]] = [(nid, 0) for nid in node_ids]

        while queue and len(result) < max_nodes:
            node_id, depth = queue.pop(0)

            if node_id in visited:
                continue

            visited.add(node_id)

            node_data = self._storage.get_graph_node(node_id)
            if node_data:
                result.append(GraphNode.from_dict(node_data))

            if depth < max_depth:
                # Get neighbors
                edges = self._storage.get_graph_edges(node_id, "both")
                for edge in edges:
                    neighbor_id = (
                        edge["target_id"]
                        if edge["source_id"] == node_id
                        else edge["source_id"]
                    )
                    if neighbor_id not in visited:
                        queue.append((neighbor_id, depth + 1))

        return result

    def find_path(
        self,
        start_id: str,
        end_id: str,
        max_depth: int = 5,
    ) -> list[GraphNode] | None:
        """
        Find a path between two nodes.

        Returns the path as a list of nodes, or None if no path exists.
        """
        visited: set[str] = set()
        queue: list[tuple[str, list[str]]] = [(start_id, [start_id])]

        while queue:
            current_id, path = queue.pop(0)

            if current_id == end_id:
                # Build node list from path
                nodes = []
                for node_id in path:
                    node_data = self._storage.get_graph_node(node_id)
                    if node_data:
                        nodes.append(GraphNode.from_dict(node_data))
                return nodes

            if len(path) >= max_depth:
                continue

            visited.add(current_id)

            edges = self._storage.get_graph_edges(current_id, "both")
            for edge in edges:
                neighbor_id = (
                    edge["target_id"]
                    if edge["source_id"] == current_id
                    else edge["source_id"]
                )
                if neighbor_id not in visited:
                    queue.append((neighbor_id, path + [neighbor_id]))

        return None

    def get_entity_context(
        self,
        entity_name: str,
        session_id: str | None = None,
        max_chunks: int = 10,
    ) -> list[GraphNode]:
        """
        Get all chunks that mention a specific entity.
        """
        # Update session if specified
        if session_id and session_id != self._session_id:
            self.set_session_id(session_id)

        # Find entity nodes matching the name
        entities = self._storage.search_graph_nodes(
            query=entity_name,
            session_id=session_id,
            node_type=NodeType.ENTITY.value,
            limit=5,
        )

        chunk_ids: set[str] = set()
        for entity in entities:
            # Find chunks that mention this entity
            edges = self._storage.get_graph_edges(
                entity["id"],
                direction="incoming",
                relationship=RelationType.MENTIONS.value,
            )
            for edge in edges:
                chunk_ids.add(edge["source_id"])

        # Get chunk nodes
        chunks: list[GraphNode] = []
        for chunk_id in list(chunk_ids)[:max_chunks]:
            node_data = self._storage.get_graph_node(chunk_id)
            if node_data:
                chunks.append(GraphNode.from_dict(node_data))

        return chunks


class GraphSearcher:
    """
    Search the knowledge graph using combined strategies.

    Combines:
    - Semantic similarity search (vector-based)
    - Keyword matching
    - Graph traversal for structural context
    """

    def __init__(self, session_id: str | None = None) -> None:
        self._session_id = session_id
        self._storage = get_graph_storage(session_id)
        self._traverser = GraphTraverser(session_id)
        self._embedding_manager = None

    def set_session_id(self, session_id: str) -> None:
        """Set the session ID and update storage."""
        self._session_id = session_id
        self._storage = get_graph_storage(session_id)
        self._traverser.set_session_id(session_id)

    def _get_embedding_manager(self):
        """Lazy load embedding manager."""
        if self._embedding_manager is None:
            try:
                from pinkyclawd.rlm.embedding import get_embedding_manager
                self._embedding_manager = get_embedding_manager()
            except ImportError:
                logger.warning("Embedding module not available")
        return self._embedding_manager

    async def search(
        self,
        query: str,
        session_id: str | None = None,
        limit: int = 10,
        expand_context: bool = True,
    ) -> list[tuple[GraphNode, float]]:
        """
        Search the graph for relevant nodes.

        Returns:
            List of (node, score) tuples sorted by relevance
        """
        # Update session if specified
        if session_id and session_id != self._session_id:
            self.set_session_id(session_id)

        results: dict[str, tuple[GraphNode, float]] = {}

        # 1. Keyword search
        keyword_matches = self._storage.search_graph_nodes(
            query=query,
            session_id=session_id,
            node_type=NodeType.CHUNK.value,
            limit=limit * 2,
        )

        for match in keyword_matches:
            node = GraphNode.from_dict(match)
            # Score based on content match (simple heuristic)
            query_terms = set(query.lower().split())
            content_terms = set(match["content"].lower().split())
            overlap = len(query_terms & content_terms)
            score = overlap / max(len(query_terms), 1)
            results[node.id] = (node, score * 0.4)  # Keyword weight: 0.4

        # 2. Semantic search
        embedding_manager = self._get_embedding_manager()
        if embedding_manager:
            try:
                from pinkyclawd.rlm.embedding import cosine_similarity

                query_embedding = await embedding_manager.embed(query)

                # Get chunks with embeddings
                chunks = self._storage.get_graph_nodes(
                    session_id=session_id,
                    node_type=NodeType.CHUNK.value,
                    limit=100,
                )

                for chunk in chunks:
                    if chunk.get("embedding"):
                        similarity = cosine_similarity(
                            query_embedding, chunk["embedding"]
                        )
                        if similarity > 0.3:  # Threshold
                            node = GraphNode.from_dict(chunk)
                            if node.id in results:
                                # Add to existing score
                                existing = results[node.id]
                                results[node.id] = (
                                    existing[0],
                                    existing[1] + similarity * 0.6,
                                )
                            else:
                                results[node.id] = (node, similarity * 0.6)

            except Exception as e:
                logger.warning(f"Semantic search failed: {e}")

        # 3. Expand context via graph traversal
        if expand_context and results:
            top_ids = [
                node.id
                for node, _ in sorted(
                    results.values(), key=lambda x: x[1], reverse=True
                )[:3]
            ]

            expanded = self._traverser.expand_context(
                node_ids=top_ids,
                max_depth=1,
                max_nodes=5,
            )

            for node in expanded:
                if node.id not in results:
                    # Add with reduced score
                    results[node.id] = (node, 0.2)

        # Sort by score and return
        sorted_results = sorted(
            results.values(), key=lambda x: x[1], reverse=True
        )
        return sorted_results[:limit]

    def search_sync(
        self,
        query: str,
        session_id: str | None = None,
        limit: int = 10,
    ) -> list[tuple[GraphNode, float]]:
        """
        Synchronous search using keyword matching only.
        """
        # Update session if specified
        if session_id and session_id != self._session_id:
            self.set_session_id(session_id)

        results: list[tuple[GraphNode, float]] = []

        # Keyword search
        matches = self._storage.search_graph_nodes(
            query=query,
            session_id=session_id,
            node_type=NodeType.CHUNK.value,
            limit=limit,
        )

        for match in matches:
            node = GraphNode.from_dict(match)
            query_terms = set(query.lower().split())
            content_terms = set(match["content"].lower().split())
            overlap = len(query_terms & content_terms)
            score = overlap / max(len(query_terms), 1)
            results.append((node, score))

        results.sort(key=lambda x: x[1], reverse=True)
        return results


# Global instances
_ingester: GraphIngester | None = None
_traverser: GraphTraverser | None = None
_searcher: GraphSearcher | None = None


def get_graph_ingester() -> GraphIngester:
    """Get the global graph ingester instance."""
    global _ingester
    if _ingester is None:
        _ingester = GraphIngester()
    return _ingester


def get_graph_traverser() -> GraphTraverser:
    """Get the global graph traverser instance."""
    global _traverser
    if _traverser is None:
        _traverser = GraphTraverser()
    return _traverser


def get_graph_searcher() -> GraphSearcher:
    """Get the global graph searcher instance."""
    global _searcher
    if _searcher is None:
        _searcher = GraphSearcher()
    return _searcher



"""
Session sharing functionality.

Generates shareable URLs and handles session sharing configuration.
"""

from __future__ import annotations

import json
import uuid
import hashlib
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Literal

from pinkyclawd.config.settings import get_config
from pinkyclawd.config.storage import Session, Message, get_storage

logger = logging.getLogger(__name__)


@dataclass
class ShareConfig:
    """Configuration for a shared session."""

    share_id: str
    session_id: str
    created_at: datetime
    expires_at: datetime | None = None
    access_level: Literal["read", "fork"] = "read"
    password_hash: str | None = None
    view_count: int = 0
    max_views: int | None = None

    def is_expired(self) -> bool:
        if self.expires_at is None:
            return False
        return datetime.now() > self.expires_at

    def is_view_limited(self) -> bool:
        if self.max_views is None:
            return False
        return self.view_count >= self.max_views

    def to_dict(self) -> dict[str, Any]:
        return {
            "share_id": self.share_id,
            "session_id": self.session_id,
            "created_at": self.created_at.isoformat(),
            "expires_at": self.expires_at.isoformat() if self.expires_at else None,
            "access_level": self.access_level,
            "view_count": self.view_count,
            "max_views": self.max_views,
        }


@dataclass
class ShareResult:
    """Result of creating a share link."""

    success: bool
    url: str | None = None
    share_id: str | None = None
    error: str | None = None


class SessionSharer:
    """
    Handles session sharing operations.

    Creates shareable links and manages share configurations.
    """

    def __init__(self) -> None:
        self._storage = get_storage()
        self._shares: dict[str, ShareConfig] = {}

    def create_share(
        self,
        session_id: str,
        expires_in: timedelta | None = None,
        access_level: Literal["read", "fork"] = "read",
        password: str | None = None,
        max_views: int | None = None,
    ) -> ShareResult:
        """
        Create a shareable link for a session.

        Args:
            session_id: Session to share
            expires_in: Time until link expires
            access_level: Read-only or allow forking
            password: Optional password protection
            max_views: Optional view limit

        Returns:
            ShareResult with URL
        """
        config = get_config()

        # Check if sharing is enabled
        if config.share == "disabled":
            return ShareResult(
                success=False,
                error="Session sharing is disabled in configuration",
            )

        # Verify session exists
        session = self._storage.get_session(session_id)
        if not session:
            return ShareResult(
                success=False,
                error=f"Session {session_id} not found",
            )

        # Generate share ID
        share_id = self._generate_share_id(session_id)

        # Calculate expiration
        expires_at = None
        if expires_in:
            expires_at = datetime.now() + expires_in

        # Hash password if provided
        password_hash = None
        if password:
            password_hash = self._hash_password(password)

        # Create share config
        share_config = ShareConfig(
            share_id=share_id,
            session_id=session_id,
            created_at=datetime.now(),
            expires_at=expires_at,
            access_level=access_level,
            password_hash=password_hash,
            max_views=max_views,
        )

        self._shares[share_id] = share_config

        # Store in session metadata
        session.metadata["share_id"] = share_id
        session.metadata["shared_at"] = datetime.now().isoformat()
        self._storage.update_session(session)

        # Generate URL
        base_url = config.server.get("share_base_url", "https://pinkyclawd.dev/share")
        url = f"{base_url}/{share_id}"

        logger.info(f"Created share link for session {session_id}: {share_id}")

        return ShareResult(
            success=True,
            url=url,
            share_id=share_id,
        )

    def get_shared_session(
        self,
        share_id: str,
        password: str | None = None,
    ) -> tuple[Session | None, list[Message] | None, str | None]:
        """
        Get a shared session by share ID.

        Args:
            share_id: Share ID from URL
            password: Password if required

        Returns:
            Tuple of (session, messages, error)
        """
        share_config = self._shares.get(share_id)
        if not share_config:
            return None, None, "Share link not found or expired"

        # Check expiration
        if share_config.is_expired():
            return None, None, "Share link has expired"

        # Check view limit
        if share_config.is_view_limited():
            return None, None, "Share link view limit reached"

        # Check password
        if share_config.password_hash:
            if not password:
                return None, None, "Password required"
            if self._hash_password(password) != share_config.password_hash:
                return None, None, "Invalid password"

        # Get session and messages
        session = self._storage.get_session(share_config.session_id)
        if not session:
            return None, None, "Session no longer exists"

        messages = self._storage.get_messages(share_config.session_id)

        # Increment view count
        share_config.view_count += 1

        return session, messages, None

    def revoke_share(self, share_id: str) -> bool:
        """Revoke a share link."""
        if share_id in self._shares:
            config = self._shares.pop(share_id)

            # Remove from session metadata
            session = self._storage.get_session(config.session_id)
            if session:
                session.metadata.pop("share_id", None)
                session.metadata.pop("shared_at", None)
                self._storage.update_session(session)

            logger.info(f"Revoked share link {share_id}")
            return True

        return False

    def list_shares(self, session_id: str | None = None) -> list[ShareConfig]:
        """List all shares, optionally filtered by session."""
        shares = list(self._shares.values())
        if session_id:
            shares = [s for s in shares if s.session_id == session_id]
        return shares

    def _generate_share_id(self, session_id: str) -> str:
        """Generate a unique share ID."""
        unique = f"{session_id}:{uuid.uuid4().hex}:{datetime.now().isoformat()}"
        hash_val = hashlib.sha256(unique.encode()).hexdigest()
        return hash_val[:12]

    def _hash_password(self, password: str) -> str:
        """Hash a password for storage."""
        return hashlib.sha256(password.encode()).hexdigest()


# Global sharer instance
_sharer: SessionSharer | None = None


def get_sharer() -> SessionSharer:
    """Get the global session sharer."""
    global _sharer
    if _sharer is None:
        _sharer = SessionSharer()
    return _sharer


def create_share_url(
    session_id: str,
    expires_in_hours: int | None = None,
) -> ShareResult:
    """
    Create a shareable URL for a session.

    Args:
        session_id: Session to share
        expires_in_hours: Hours until expiration

    Returns:
        ShareResult with URL
    """
    expires_in = None
    if expires_in_hours:
        expires_in = timedelta(hours=expires_in_hours)

    return get_sharer().create_share(session_id, expires_in=expires_in)



"""
LSP tool for code intelligence operations.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

from pinkyclawd.tool.base import (
    INTEGER_PARAM,
    STRING_PARAM,
    Tool,
    ToolContext,
    ToolResult,
    make_schema,
)


class LSPTool(Tool):
    """Language Server Protocol operations for code intelligence."""

    @property
    def name(self) -> str:
        return "lsp"

    @property
    def description(self) -> str:
        return """Interact with Language Server Protocol (LSP) for code intelligence.

Supported operations:
- goToDefinition: Find where a symbol is defined
- findReferences: Find all references to a symbol
- hover: Get documentation/type info for a symbol
- documentSymbol: Get all symbols in a file
- workspaceSymbol: Search for symbols across the workspace
- goToImplementation: Find implementations of interface/abstract
- prepareCallHierarchy: Get call hierarchy at position
- incomingCalls: Find callers of a function
- outgoingCalls: Find functions called by a function

Line and character positions are 1-indexed (as shown in editors)."""

    @property
    def parameters(self) -> dict[str, Any]:
        return make_schema(
            required=[
                (
                    "operation",
                    "LSP operation to perform",
                    {
                        "type": "string",
                        "enum": [
                            "goToDefinition",
                            "findReferences",
                            "hover",
                            "documentSymbol",
                            "workspaceSymbol",
                            "goToImplementation",
                            "prepareCallHierarchy",
                            "incomingCalls",
                            "outgoingCalls",
                        ],
                    },
                ),
                ("filePath", "Path to the file", STRING_PARAM),
                ("line", "Line number (1-indexed)", INTEGER_PARAM),
                ("character", "Character offset (1-indexed)", INTEGER_PARAM),
            ],
            optional=[
                (
                    "query",
                    "Search query (for workspaceSymbol operation)",
                    STRING_PARAM,
                ),
            ],
        )

    async def execute(self, ctx: ToolContext, **kwargs: Any) -> ToolResult:
        """Execute an LSP operation."""
        operation = kwargs.get("operation", "")
        file_path = kwargs.get("filePath", "")
        line = kwargs.get("line", 1)
        character = kwargs.get("character", 1)
        query = kwargs.get("query", "")

        if not operation:
            return ToolResult.fail("No operation specified")

        if not file_path:
            return ToolResult.fail("No file path specified")

        # Resolve path
        path = Path(file_path)
        if not path.is_absolute():
            path = ctx.working_directory / path

        if not path.exists():
            return ToolResult.fail(f"File not found: {path}")

        try:
            # Import here to avoid circular imports and allow lazy loading
            from pinkyclawd.lsp import (
                document_symbol,
                find_references,
                get_call_hierarchy,
                get_incoming_calls,
                get_lsp_client,
                get_outgoing_calls,
                go_to_definition,
                go_to_implementation,
                hover,
                workspace_symbol,
            )

            # Ensure server is running
            client = get_lsp_client()
            state = await client.ensure_server_for_file(path, ctx.working_directory)
            if not state:
                return ToolResult.fail(
                    f"No language server available for {path.suffix} files. "
                    "The server may need to be installed."
                )

            if operation == "goToDefinition":
                locations = await go_to_definition(path, line, character)
                return self._format_locations(locations, "definitions")

            elif operation == "findReferences":
                locations = await find_references(path, line, character)
                return self._format_locations(locations, "references")

            elif operation == "hover":
                result = await hover(path, line, character)
                if not result:
                    return ToolResult.ok("No hover information available")
                return ToolResult.ok(
                    result.contents,
                    has_range=result.range is not None,
                )

            elif operation == "documentSymbol":
                symbols = await document_symbol(path)
                return self._format_symbols(symbols)

            elif operation == "workspaceSymbol":
                if not query:
                    return ToolResult.fail("Query required for workspaceSymbol")
                symbols = await workspace_symbol(query, path)
                return self._format_workspace_symbols(symbols)

            elif operation == "goToImplementation":
                locations = await go_to_implementation(path, line, character)
                return self._format_locations(locations, "implementations")

            elif operation == "prepareCallHierarchy":
                items = await get_call_hierarchy(path, line, character)
                if not items:
                    return ToolResult.ok("No call hierarchy items found")
                lines = []
                for item in items:
                    lines.append(f"{item.name} ({item.detail or 'no detail'})")
                    lines.append(f"  {item.uri}:{item.range.start.line + 1}")
                return ToolResult.ok("\n".join(lines), count=len(items))

            elif operation == "incomingCalls":
                calls = await get_incoming_calls(path, line, character)
                if not calls:
                    return ToolResult.ok("No incoming calls found")
                lines = []
                for call in calls:
                    item = call.item
                    lines.append(f"Called by: {item.name}")
                    lines.append(f"  {item.uri}:{item.range.start.line + 1}")
                return ToolResult.ok("\n".join(lines), count=len(calls))

            elif operation == "outgoingCalls":
                calls = await get_outgoing_calls(path, line, character)
                if not calls:
                    return ToolResult.ok("No outgoing calls found")
                lines = []
                for call in calls:
                    item = call.item
                    lines.append(f"Calls: {item.name}")
                    lines.append(f"  {item.uri}:{item.range.start.line + 1}")
                return ToolResult.ok("\n".join(lines), count=len(calls))

            else:
                return ToolResult.fail(f"Unknown operation: {operation}")

        except Exception as e:
            return ToolResult.fail(f"LSP operation failed: {e}")

    def _format_locations(
        self,
        locations: list,
        label: str,
    ) -> ToolResult:
        """Format location results."""
        if not locations:
            return ToolResult.ok(f"No {label} found")

        lines = [f"Found {len(locations)} {label}:"]
        for loc in locations:
            path = loc.file_path
            line = loc.range.start.line + 1
            char = loc.range.start.character + 1
            lines.append(f"  {path}:{line}:{char}")

        return ToolResult.ok("\n".join(lines), count=len(locations))

    def _format_symbols(self, symbols: list) -> ToolResult:
        """Format document symbols."""
        if not symbols:
            return ToolResult.ok("No symbols found")

        from pinkyclawd.lsp.operations import get_symbol_kind_name

        lines = [f"Found {len(symbols)} symbols:"]

        def format_symbol(sym, indent: int = 0) -> None:
            prefix = "  " * indent
            kind = get_symbol_kind_name(sym.kind)
            detail = f" - {sym.detail}" if sym.detail else ""
            line = sym.range.start.line + 1
            lines.append(f"{prefix}{kind}: {sym.name}{detail} (line {line})")

            if sym.children:
                for child in sym.children:
                    format_symbol(child, indent + 1)

        for sym in symbols:
            format_symbol(sym)

        return ToolResult.ok("\n".join(lines), count=len(symbols))

    def _format_workspace_symbols(self, symbols: list) -> ToolResult:
        """Format workspace symbols."""
        if not symbols:
            return ToolResult.ok("No symbols found")

        from pinkyclawd.lsp.operations import get_symbol_kind_name

        lines = [f"Found {len(symbols)} symbols:"]
        for sym in symbols:
            kind = get_symbol_kind_name(sym.kind)
            container = f" in {sym.container_name}" if sym.container_name else ""
            path = sym.location.file_path
            line = sym.location.range.start.line + 1
            lines.append(f"  {kind}: {sym.name}{container}")
            lines.append(f"    {path}:{line}")

        return ToolResult.ok("\n".join(lines), count=len(symbols))



"""
Edit tool for search/replace operations on files.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

from pinkyclawd.tool.base import (
    Tool,
    ToolContext,
    ToolResult,
    make_schema,
    STRING_PARAM,
    BOOLEAN_PARAM,
)


class EditTool(Tool):
    """Edit files using exact string replacement."""

    @property
    def name(self) -> str:
        return "edit"

    @property
    def description(self) -> str:
        return """Perform exact string replacements in files.

You MUST use the Read tool before editing to see exact content.
The oldString must match exactly (including whitespace/indentation).
The edit will FAIL if oldString is not found.
The edit will FAIL if oldString matches multiple times (provide more context).
Use replaceAll=true to replace all occurrences."""

    @property
    def parameters(self) -> dict[str, Any]:
        return make_schema(
            required=[
                ("filePath", "Absolute path to the file to modify", STRING_PARAM),
                ("oldString", "The exact text to replace", STRING_PARAM),
                (
                    "newString",
                    "The text to replace it with (must differ from oldString)",
                    STRING_PARAM,
                ),
            ],
            optional=[
                (
                    "replaceAll",
                    "Replace all occurrences (default: false)",
                    BOOLEAN_PARAM,
                ),
            ],
        )

    @property
    def requires_confirmation(self) -> bool:
        return True

    async def execute(self, ctx: ToolContext, **kwargs: Any) -> ToolResult:
        """Edit a file."""
        file_path = kwargs.get("filePath", "")
        old_string = kwargs.get("oldString", "")
        new_string = kwargs.get("newString", "")
        replace_all = kwargs.get("replaceAll", False)

        if not file_path:
            return ToolResult.fail("No file path provided")
        if not old_string:
            return ToolResult.fail("No oldString provided")
        if old_string == new_string:
            return ToolResult.fail("oldString and newString are identical")

        path = Path(file_path)
        if not path.is_absolute():
            path = ctx.working_directory / path

        if not path.exists():
            return ToolResult.fail(f"File not found: {path}")

        try:
            content = path.read_text(encoding="utf-8")
        except Exception as e:
            return ToolResult.fail(f"Failed to read file: {e}")

        # Count occurrences
        count = content.count(old_string)

        if count == 0:
            return ToolResult.fail("oldString not found in content")

        if count > 1 and not replace_all:
            return ToolResult.fail(
                f"oldString found {count} times. "
                "Provide more context to uniquely identify the match, "
                "or use replaceAll=true to replace all occurrences."
            )

        # Perform replacement
        if replace_all:
            new_content = content.replace(old_string, new_string)
        else:
            new_content = content.replace(old_string, new_string, 1)

        try:
            path.write_text(new_content, encoding="utf-8")
        except Exception as e:
            return ToolResult.fail(f"Failed to write file: {e}")

        return ToolResult.ok(
            f"Replaced {count if replace_all else 1} occurrence(s) in {path}",
            replacements=count if replace_all else 1,
        )



"""
LLM-based summarization for RLM context archival.

Generates detailed summaries of archived conversations using
a compaction agent, enabling better context continuation.
"""

from __future__ import annotations

import logging
from typing import Any

from pinkyclawd.config.settings import get_config
from pinkyclawd.config.storage import Message, MessagePart, PartType
from pinkyclawd.rlm.display import get_console

logger = logging.getLogger(__name__)


# Prompt for generating context summaries
SUMMARY_PROMPT = """Provide a detailed summary of this conversation for context continuation. Focus on:

1. **What was accomplished**: Files modified, features implemented, bugs fixed
2. **Current state**: Where the work stands, any pending items
3. **Key decisions**: Important choices made and their rationale
4. **Important locations**: File paths, function names, code locations mentioned
5. **Errors & resolutions**: Any issues encountered and how they were resolved

Keep the summary concise but comprehensive enough to continue the conversation in a new context window."""


async def generate_summary(
    messages: list[Message],
    model: str | None = None,
) -> str:
    """
    Generate an LLM-based summary of conversation messages.

    Uses a lightweight model to create a detailed summary suitable
    for context continuation.

    Args:
        messages: Messages to summarize
        model: Optional model override (defaults to config model)

    Returns:
        Generated summary text
    """
    if not messages:
        return "No messages to summarize"

    config = get_config()
    model = model or config.model

    # Build conversation text for summarization
    conversation_text = _serialize_messages(messages)

    if not conversation_text.strip():
        return "No content to summarize"

    console = get_console()
    console.print(
        f"[bold yellow]RLM[/bold yellow] [dim]Generating context summary...[/dim]"
    )

    try:
        # Try to use provider for summarization
        summary = await _call_llm_for_summary(conversation_text, model)

        if summary:
            logger.info(f"Generated summary: {len(summary)} chars")
            console.print(
                f"[bold green]RLM[/bold green] [dim]Summary generated ({len(summary)} chars)[/dim]"
            )
            return summary

    except Exception as e:
        logger.warning(f"LLM summarization failed: {e}, using fallback")

    # Fallback to extractive summary
    return _generate_fallback_summary(messages)


async def _call_llm_for_summary(conversation_text: str, model: str) -> str | None:
    """
    Call an LLM to generate a summary.

    Args:
        conversation_text: Serialized conversation
        model: Model to use

    Returns:
        Generated summary or None if failed
    """
    try:
        from pinkyclawd.provider.registry import get_provider_registry

        registry = get_provider_registry()
        provider = registry.get_provider_for_model(model)

        if not provider:
            logger.warning(f"No provider found for model {model}")
            return None

        # Create a simple completion request
        messages = [
            {
                "role": "system",
                "content": "You are a helpful assistant that summarizes technical conversations concisely.",
            },
            {
                "role": "user",
                "content": f"{SUMMARY_PROMPT}\n\n---\n\n{conversation_text}",
            },
        ]

        # Use a small response for efficiency
        response = await provider.complete(
            messages=messages,
            model=model,
            max_tokens=1000,
        )

        return response.get("content", "").strip()

    except ImportError:
        logger.debug("Provider registry not available for summarization")
        return None
    except Exception as e:
        logger.warning(f"LLM call failed: {e}")
        return None


def _generate_fallback_summary(messages: list[Message]) -> str:
    """
    Generate a fallback extractive summary when LLM is unavailable.

    Extracts key information from messages without using an LLM.

    Args:
        messages: Messages to summarize

    Returns:
        Extractive summary text
    """
    parts = []

    # Find key information
    files_mentioned = set()
    tools_used = set()
    errors_found = []

    for msg in messages:
        for part in msg.parts:
            if part.type == PartType.TEXT:
                text = part.content.get("text", "")

                # Extract file paths
                import re
                file_matches = re.findall(r'[./\w]+\.[a-z]{1,4}', text)
                files_mentioned.update(file_matches[:10])

            elif part.type == PartType.TOOL_USE:
                tool_name = part.content.get("name", "")
                if tool_name:
                    tools_used.add(tool_name)

            elif part.type == PartType.TOOL_RESULT:
                result = part.content.get("result", "")
                if "error" in result.lower() or "failed" in result.lower():
                    errors_found.append(result[:200])

    # Build summary
    parts.append("## Conversation Summary (Auto-generated)")
    parts.append("")

    if tools_used:
        parts.append(f"**Tools used**: {', '.join(sorted(tools_used))}")

    if files_mentioned:
        parts.append(f"**Files involved**: {', '.join(sorted(list(files_mentioned)[:10]))}")

    if errors_found:
        parts.append(f"**Errors encountered**: {len(errors_found)}")

    parts.append("")
    parts.append(f"**Message count**: {len(messages)}")

    # Add first and last user messages as context
    user_messages = [m for m in messages if m.role.value == "user"]
    if user_messages:
        first_user = _get_text_from_message(user_messages[0])
        if first_user:
            parts.append(f"\n**Started with**: {first_user[:200]}...")

        if len(user_messages) > 1:
            last_user = _get_text_from_message(user_messages[-1])
            if last_user:
                parts.append(f"\n**Ended with**: {last_user[:200]}...")

    return "\n".join(parts)


def _serialize_messages(messages: list[Message]) -> str:
    """
    Serialize messages to text format for summarization.

    Args:
        messages: Messages to serialize

    Returns:
        Serialized text
    """
    parts = []

    for msg in messages:
        role = msg.role.value.upper()
        parts.append(f"[{role}]")

        for part in msg.parts:
            if part.type == PartType.TEXT:
                text = part.content.get("text", "")
                if text:
                    # Truncate very long text
                    if len(text) > 1000:
                        text = text[:1000] + "...[truncated]"
                    parts.append(text)

            elif part.type == PartType.TOOL_USE:
                tool_name = part.content.get("name", "unknown")
                parts.append(f"[Tool: {tool_name}]")

            elif part.type == PartType.TOOL_RESULT:
                result = part.content.get("result", "")
                if len(result) > 500:
                    result = result[:500] + "...[truncated]"
                parts.append(f"[Result: {result}]")

        parts.append("")

    return "\n".join(parts)


def _get_text_from_message(message: Message) -> str:
    """Extract text content from a message."""
    for part in message.parts:
        if part.type == PartType.TEXT:
            return part.content.get("text", "")
    return ""


def generate_task_summary(
    task_description: str,
    messages: list[Message],
) -> str:
    """
    Generate a quick summary for task-based archival.

    Synchronous version that doesn't use LLM, suitable for
    immediate task completion archival.

    Args:
        task_description: Description of the completed task
        messages: Messages related to the task

    Returns:
        Task summary text
    """
    parts = [f"Task: {task_description}"]

    # Count what was done
    tool_uses = 0
    files_modified = set()

    for msg in messages:
        for part in msg.parts:
            if part.type == PartType.TOOL_USE:
                tool_uses += 1
                tool_name = part.content.get("name", "")
                args = part.content.get("arguments", {})

                # Track file modifications
                if tool_name in ("write", "edit"):
                    file_path = args.get("file_path", args.get("filePath", ""))
                    if file_path:
                        files_modified.add(file_path)

    if tool_uses:
        parts.append(f"Tool executions: {tool_uses}")

    if files_modified:
        parts.append(f"Files modified: {', '.join(sorted(files_modified))}")

    parts.append(f"Messages: {len(messages)}")

    return " | ".join(parts)



"""
Task tool for launching sub-agents.
"""

from __future__ import annotations

from typing import Any

from pinkyclawd.tool.base import Tool, ToolContext, ToolResult, make_schema, STRING_PARAM


class TaskTool(Tool):
    """Launch sub-agents for parallel work."""

    @property
    def name(self) -> str:
        return "task"

    @property
    def description(self) -> str:
        return """Launch a sub-agent to handle complex, multistep tasks autonomously.

Available agent types:
- general: For researching and executing multi-step tasks in parallel
- explore: Fast agent for codebase exploration (find files, search code)

Use this when:
- Task requires multiple independent pieces of work
- You need to explore a codebase without using context
- Task matches a specialized agent's description

Launch multiple agents in parallel when possible for efficiency."""

    @property
    def parameters(self) -> dict[str, Any]:
        return make_schema(
            required=[
                (
                    "description",
                    "Short (3-5 words) description of the task",
                    STRING_PARAM,
                ),
                ("prompt", "Detailed task for the agent to perform", STRING_PARAM),
                (
                    "subagent_type",
                    "Type of agent to use",
                    {
                        "type": "string",
                        "enum": ["general", "explore"],
                    },
                ),
            ],
            optional=[
                ("session_id", "Existing task session to continue", STRING_PARAM),
                ("command", "The command that triggered this task", STRING_PARAM),
            ],
        )

    async def execute(self, ctx: ToolContext, **kwargs: Any) -> ToolResult:
        """Launch a sub-agent task."""
        description = kwargs.get("description", "")
        prompt = kwargs.get("prompt", "")
        subagent_type = kwargs.get("subagent_type", "general")

        if not description:
            return ToolResult.fail("No description provided")
        if not prompt:
            return ToolResult.fail("No prompt provided")

        # In the full implementation, this would spawn a sub-agent
        # For now, return task info for the orchestrator to handle
        return ToolResult.ok(
            f"Launching {subagent_type} agent: {description}",
            description=description,
            prompt=prompt,
            subagent_type=subagent_type,
            task_launched=True,
        )



"""
Context retrieval system for RLM.

Proactively retrieves relevant archived context to inject into
the conversation, enabling coherent long-running sessions.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Any

from pinkyclawd.config.settings import get_config
from pinkyclawd.config.storage import ContextBlock, Message, PartType
from pinkyclawd.events import EventType, Event, get_event_bus
from pinkyclawd.rlm.search import ContextSearcher, SearchResult, get_searcher
from pinkyclawd.rlm.context import TokenCounter

logger = logging.getLogger(__name__)


@dataclass
class RetrievalContext:
    """Context to be injected into the conversation."""

    blocks: list[ContextBlock] = field(default_factory=list)
    total_tokens: int = 0
    trigger: str = ""  # What triggered the retrieval

    @property
    def is_empty(self) -> bool:
        return len(self.blocks) == 0

    def to_system_message(self) -> str:
        """Format as a system message for injection."""
        if self.is_empty:
            return ""

        parts = ["<archived-context>"]
        parts.append(
            f"The following is relevant context from earlier in this conversation or previous sessions:"
        )
        parts.append("")

        for block in self.blocks:
            parts.append(f"--- Context Block (Task: {block.task_description or 'General'}) ---")
            parts.append(block.content)
            parts.append("")

        parts.append("</archived-context>")
        return "\n".join(parts)

    def to_dict(self) -> dict[str, Any]:
        return {
            "blocks": [b.to_dict() for b in self.blocks],
            "total_tokens": self.total_tokens,
            "trigger": self.trigger,
        }


class ContextRetriever:
    """
    Retrieves relevant archived context for injection.

    Analyzes the current conversation and user query to determine
    what archived context might be relevant.
    """

    def __init__(self) -> None:
        self._searcher = get_searcher()
        self._counter = TokenCounter()
        self._event_bus = get_event_bus()

    def retrieve_for_query(
        self,
        query: str,
        session_id: str,
        max_tokens: int | None = None,
    ) -> RetrievalContext:
        """
        Retrieve context relevant to a user query.

        Args:
            query: User's query/message
            session_id: Current session ID
            max_tokens: Maximum tokens to retrieve

        Returns:
            RetrievalContext with relevant blocks
        """
        config = get_config()

        if not config.rlm.enabled or not config.rlm.auto_retrieve:
            return RetrievalContext()

        # Set default max tokens (10% of context limit)
        if max_tokens is None:
            max_tokens = int(self._counter.get_model_limit(config.model) * 0.1)

        # Search for relevant blocks
        results = self._searcher.search(
            query=query,
            session_id=None,  # Search across all sessions
            limit=20,
        )

        # Filter by threshold
        threshold = config.rlm.semantic_search.get("proactive_threshold", 0.4)
        relevant_results = [r for r in results if r.score >= threshold]

        if not relevant_results:
            return RetrievalContext()

        # Select blocks within token budget
        selected_blocks = self._select_within_budget(relevant_results, max_tokens)

        if not selected_blocks:
            return RetrievalContext()

        total_tokens = sum(b.tokens for b in selected_blocks)

        context = RetrievalContext(
            blocks=selected_blocks,
            total_tokens=total_tokens,
            trigger=f"query: {query[:50]}...",
        )

        # Emit event
        self._emit_retrieval_event(session_id, context)

        logger.info(
            f"Retrieved {len(selected_blocks)} context blocks ({total_tokens} tokens) for query"
        )

        return context

    def retrieve_for_task(
        self,
        task_description: str,
        session_id: str,
        max_tokens: int | None = None,
    ) -> RetrievalContext:
        """
        Retrieve context relevant to a task being worked on.

        Args:
            task_description: Description of the current task
            session_id: Current session ID
            max_tokens: Maximum tokens to retrieve

        Returns:
            RetrievalContext with relevant blocks
        """
        return self.retrieve_for_query(
            query=task_description,
            session_id=session_id,
            max_tokens=max_tokens,
        )

    def retrieve_for_message(
        self,
        message: Message,
        session_id: str,
        max_tokens: int | None = None,
    ) -> RetrievalContext:
        """
        Retrieve context based on a message's content.

        Args:
            message: Message to analyze
            session_id: Current session ID
            max_tokens: Maximum tokens to retrieve

        Returns:
            RetrievalContext with relevant blocks
        """
        # Extract text from message parts
        text_parts = []
        for part in message.parts:
            if part.type == PartType.TEXT:
                text_parts.append(part.content.get("text", ""))

        if not text_parts:
            return RetrievalContext()

        query = " ".join(text_parts)
        return self.retrieve_for_query(query, session_id, max_tokens)

    def retrieve_recent(
        self,
        session_id: str,
        limit: int = 5,
    ) -> RetrievalContext:
        """
        Retrieve most recent archived blocks for a session.

        Args:
            session_id: Session ID
            limit: Maximum blocks to retrieve

        Returns:
            RetrievalContext with recent blocks
        """
        blocks = self._searcher.get_recent(session_id=session_id, limit=limit)

        if not blocks:
            return RetrievalContext()

        total_tokens = sum(b.tokens for b in blocks)

        return RetrievalContext(
            blocks=blocks,
            total_tokens=total_tokens,
            trigger="recent",
        )

    def retrieve_by_task_id(self, task_id: str) -> RetrievalContext:
        """
        Retrieve context for a specific task.

        Args:
            task_id: Task ID to look up

        Returns:
            RetrievalContext with task's archived content
        """
        block = self._searcher.get_by_task(task_id)

        if not block:
            return RetrievalContext()

        return RetrievalContext(
            blocks=[block],
            total_tokens=block.tokens,
            trigger=f"task: {task_id}",
        )

    def _select_within_budget(
        self,
        results: list[SearchResult],
        max_tokens: int,
    ) -> list[ContextBlock]:
        """Select blocks that fit within token budget."""
        selected = []
        total_tokens = 0

        for result in results:
            if total_tokens + result.block.tokens > max_tokens:
                continue

            selected.append(result.block)
            total_tokens += result.block.tokens

        return selected

    def _emit_retrieval_event(
        self,
        session_id: str,
        context: RetrievalContext,
    ) -> None:
        """Emit context retrieved event."""
        self._event_bus.emit_sync(
            Event(
                type=EventType.RLM_CONTEXT_RETRIEVED,
                data={
                    "session_id": session_id,
                    "block_count": len(context.blocks),
                    "total_tokens": context.total_tokens,
                    "trigger": context.trigger,
                },
            )
        )


# Functions for the mcp_rlm_query tool


def search(query: str, limit: int = 10) -> list[dict]:
    """Search archived context by keywords."""
    searcher = get_searcher()
    results = searcher.search(query=query, limit=limit)
    return [r.to_dict() for r in results]


def get_recent(n: int = 10) -> list[dict]:
    """Get the N most recent context blocks."""
    searcher = get_searcher()
    blocks = searcher.get_recent(limit=n)
    return [b.to_dict() for b in blocks]


def get_by_task(task_id: str) -> dict | None:
    """Get context for a specific task ID."""
    searcher = get_searcher()
    block = searcher.get_by_task(task_id)
    return block.to_dict() if block else None


def get_by_session(session_id: str) -> list[dict]:
    """Get all blocks from a specific session."""
    searcher = get_searcher()
    blocks = searcher.get_recent(session_id=session_id, limit=1000)
    return [b.to_dict() for b in blocks]


def list_tasks() -> list[dict]:
    """List all archived tasks."""
    searcher = get_searcher()
    blocks = searcher.get_recent(limit=1000)
    tasks = []
    for block in blocks:
        if block.task_id:
            tasks.append(
                {
                    "task_id": block.task_id,
                    "description": block.task_description,
                    "session_id": block.session_id,
                    "tokens": block.tokens,
                    "created_at": block.created_at.isoformat(),
                }
            )
    return tasks


def list_sessions() -> list[dict]:
    """List all sessions with archived context."""
    searcher = get_searcher()
    blocks = searcher.get_recent(limit=1000)

    sessions: dict[str, dict] = {}
    for block in blocks:
        if block.session_id not in sessions:
            sessions[block.session_id] = {
                "session_id": block.session_id,
                "block_count": 0,
                "total_tokens": 0,
                "latest": block.created_at.isoformat(),
            }
        sessions[block.session_id]["block_count"] += 1
        sessions[block.session_id]["total_tokens"] += block.tokens

    return list(sessions.values())


# Global retriever instance
_retriever: ContextRetriever | None = None


def get_retriever() -> ContextRetriever:
    """Get the global retriever instance."""
    global _retriever
    if _retriever is None:
        _retriever = ContextRetriever()
    return _retriever



"""
Agent system for PinkyClawd.

Provides different agent personalities and capabilities for various
coding tasks like building, planning, and exploration.
"""

from pinkyclawd.agent.base import Agent, AgentMode, AgentPermissions
from pinkyclawd.agent.build import create_build_agent, get_build_agent
from pinkyclawd.agent.plan import create_plan_agent, get_plan_agent
from pinkyclawd.agent.explore import create_explore_agent, get_explore_agent
from pinkyclawd.agent.registry import (
    AgentRegistry,
    get_agent_registry,
    get_agent,
    list_agents,
    cycle_agent,
)

__all__ = [
    "Agent",
    "AgentMode",
    "AgentPermissions",
    "create_build_agent",
    "get_build_agent",
    "create_plan_agent",
    "get_plan_agent",
    "create_explore_agent",
    "get_explore_agent",
    "AgentRegistry",
    "get_agent_registry",
    "get_agent",
    "list_agents",
    "cycle_agent",
]



"""
Session manager for creating, loading, and managing sessions.

Handles session lifecycle including creation, forking, compaction,
and child session relationships.
"""

from __future__ import annotations

import uuid
import logging
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Literal

from pinkyclawd.config.storage import (
    Session,
    Message,
    MessagePart,
    MessageRole,
    PartType,
    get_storage,
)
from pinkyclawd.events import (
    EventType,
    SessionEvent,
    get_event_bus,
    emit_session_created,
)
from pinkyclawd.rlm.context import get_context_manager
from pinkyclawd.rlm.archive import get_archiver

logger = logging.getLogger(__name__)


@dataclass
class SessionInfo:
    """Summary information about a session."""

    id: str
    title: str
    directory: str
    message_count: int
    created_at: datetime
    updated_at: datetime
    is_archived: bool
    parent_id: str | None = None
    child_count: int = 0

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "title": self.title,
            "directory": self.directory,
            "message_count": self.message_count,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "is_archived": self.is_archived,
            "parent_id": self.parent_id,
            "child_count": self.child_count,
        }


class SessionManager:
    """
    Manages session lifecycle and operations.

    Provides CRUD operations, forking, compaction, and session relationships.
    """

    _instance: SessionManager | None = None

    def __new__(cls) -> SessionManager:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self) -> None:
        if self._initialized:
            return

        self._storage = get_storage()
        self._event_bus = get_event_bus()
        self._context_manager = get_context_manager()
        self._archiver = get_archiver()
        self._current_session_id: str | None = None
        self._initialized = True

    @property
    def current_session_id(self) -> str | None:
        """Get the current active session ID."""
        return self._current_session_id

    def create(
        self,
        title: str = "New Session",
        directory: str | Path | None = None,
        parent_id: str | None = None,
    ) -> Session:
        """
        Create a new session.

        Args:
            title: Session title
            directory: Working directory for the session
            parent_id: Optional parent session for forking

        Returns:
            Created session
        """
        session_id = f"ses_{uuid.uuid4().hex[:12]}"

        session = Session(
            id=session_id,
            title=title,
            directory=str(directory) if directory else str(Path.cwd()),
            parent_id=parent_id,
            created_at=datetime.now(),
            updated_at=datetime.now(),
        )

        self._storage.create_session(session)
        self._current_session_id = session_id

        # Emit event
        emit_session_created(session_id, title=title)

        logger.info(f"Created session {session_id}: {title}")
        return session

    def get(self, session_id: str) -> Session | None:
        """Get a session by ID."""
        return self._storage.get_session(session_id)

    def get_or_create(
        self,
        session_id: str | None = None,
        directory: str | Path | None = None,
    ) -> Session:
        """Get an existing session or create a new one."""
        if session_id:
            session = self.get(session_id)
            if session:
                self._current_session_id = session_id
                return session

        return self.create(directory=directory)

    def list(
        self,
        limit: int = 50,
        offset: int = 0,
        directory: str | None = None,
    ) -> list[SessionInfo]:
        """
        List sessions with summary information.

        Args:
            limit: Maximum sessions to return
            offset: Pagination offset
            directory: Filter by directory

        Returns:
            List of session info objects
        """
        sessions = self._storage.list_sessions(
            limit=limit,
            offset=offset,
            directory=directory,
        )

        result = []
        for session in sessions:
            messages = self._storage.get_messages(session.id)
            children = self._get_children(session.id)

            result.append(
                SessionInfo(
                    id=session.id,
                    title=session.title,
                    directory=session.directory,
                    message_count=len(messages),
                    created_at=session.created_at,
                    updated_at=session.updated_at,
                    is_archived=session.archived_at is not None,
                    parent_id=session.parent_id,
                    child_count=len(children),
                )
            )

        return result

    def update(
        self,
        session_id: str,
        title: str | None = None,
        directory: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> Session | None:
        """Update session properties."""
        session = self._storage.get_session(session_id)
        if not session:
            return None

        if title is not None:
            session.title = title
        if directory is not None:
            session.directory = directory
        if metadata is not None:
            session.metadata.update(metadata)

        self._storage.update_session(session)

        # Emit event
        self._event_bus.emit_sync(
            SessionEvent(
                type=EventType.SESSION_UPDATED,
                data={"session_id": session_id},
            )
        )

        return session

    def delete(self, session_id: str) -> bool:
        """
        Delete a session and all its data.

        Args:
            session_id: Session to delete

        Returns:
            True if deleted
        """
        session = self._storage.get_session(session_id)
        if not session:
            return False

        # Delete child sessions first
        children = self._get_children(session_id)
        for child in children:
            self.delete(child.id)

        self._storage.delete_session(session_id)
        self._context_manager.reset_session(session_id)

        # Emit event
        self._event_bus.emit_sync(
            SessionEvent(
                type=EventType.SESSION_DELETED,
                data={"session_id": session_id},
            )
        )

        # Clear current session if deleted
        if self._current_session_id == session_id:
            self._current_session_id = None

        logger.info(f"Deleted session {session_id}")
        return True

    def fork(
        self,
        session_id: str,
        from_message_id: str | None = None,
        title: str | None = None,
    ) -> Session | None:
        """
        Fork a session from a specific point.

        Args:
            session_id: Session to fork
            from_message_id: Message to fork from (includes this message)
            title: Title for new session

        Returns:
            New forked session
        """
        parent = self._storage.get_session(session_id)
        if not parent:
            return None

        messages = self._storage.get_messages(session_id)

        # Find messages up to fork point
        if from_message_id:
            fork_messages = []
            for msg in messages:
                fork_messages.append(msg)
                if msg.id == from_message_id:
                    break
        else:
            fork_messages = messages

        # Create new session
        fork_title = title or f"Fork of {parent.title}"
        new_session = self.create(
            title=fork_title,
            directory=parent.directory,
            parent_id=session_id,
        )

        # Copy messages
        for msg in fork_messages:
            new_msg = Message(
                id=f"msg_{uuid.uuid4().hex[:12]}",
                session_id=new_session.id,
                role=msg.role,
                parts=[
                    MessagePart(
                        id=f"part_{uuid.uuid4().hex[:12]}",
                        message_id="",  # Will be set below
                        type=part.type,
                        content=part.content.copy(),
                    )
                    for part in msg.parts
                ],
                metadata=msg.metadata.copy(),
            )
            # Fix message IDs in parts
            for part in new_msg.parts:
                part.message_id = new_msg.id

            self._storage.add_message(new_msg)

        logger.info(
            f"Forked session {session_id} to {new_session.id} with {len(fork_messages)} messages"
        )

        return new_session

    def compact(self, session_id: str) -> bool:
        """
        Compact a session by archiving old messages.

        Archives messages and removes them from active context.

        Args:
            session_id: Session to compact

        Returns:
            True if compaction performed
        """
        session = self._storage.get_session(session_id)
        if not session:
            return False

        # Emit compacting event
        self._event_bus.emit_sync(
            SessionEvent(
                type=EventType.SESSION_COMPACTING,
                data={"session_id": session_id},
            )
        )

        # Archive oldest messages
        result = self._archiver.archive_oldest(session_id)

        if result.success:
            logger.info(
                f"Compacted session {session_id}: "
                f"archived {result.messages_archived} messages "
                f"({result.tokens_archived} tokens)"
            )
            return True

        return False

    def rename(self, session_id: str, new_title: str) -> bool:
        """Rename a session."""
        session = self.update(session_id, title=new_title)
        return session is not None

    def archive(self, session_id: str) -> bool:
        """Mark a session as archived."""
        session = self._storage.get_session(session_id)
        if not session:
            return False

        session.archived_at = datetime.now()
        self._storage.update_session(session)
        return True

    def unarchive(self, session_id: str) -> bool:
        """Unarchive a session."""
        session = self._storage.get_session(session_id)
        if not session:
            return False

        session.archived_at = None
        self._storage.update_session(session)
        return True

    def get_messages(self, session_id: str) -> list[Message]:
        """Get all messages for a session."""
        return self._storage.get_messages(session_id)

    def add_message(self, session_id: str, message: Message) -> Message:
        """Add a message to a session."""
        message.session_id = session_id
        self._storage.add_message(message)

        # Update context manager
        self._context_manager.add_message(session_id, message)

        # Update session timestamp
        session = self._storage.get_session(session_id)
        if session:
            session.updated_at = datetime.now()
            self._storage.update_session(session)

        return message

    def set_current(self, session_id: str) -> bool:
        """Set the current active session."""
        session = self._storage.get_session(session_id)
        if session:
            self._current_session_id = session_id
            return True
        return False

    def _get_children(self, session_id: str) -> list[Session]:
        """Get child sessions of a parent."""
        all_sessions = self._storage.list_sessions(limit=1000)
        return [s for s in all_sessions if s.parent_id == session_id]

    @classmethod
    def reset(cls) -> None:
        """Reset singleton (for testing)."""
        cls._instance = None


def get_session_manager() -> SessionManager:
    """Get the global session manager."""
    return SessionManager()


# Convenience functions


def create_session(
    title: str = "New Session",
    directory: str | Path | None = None,
) -> Session:
    """Create a new session."""
    return get_session_manager().create(title=title, directory=directory)


def get_session(session_id: str) -> Session | None:
    """Get a session by ID."""
    return get_session_manager().get(session_id)


def list_sessions(
    limit: int = 50,
    directory: str | None = None,
) -> list[SessionInfo]:
    """List sessions."""
    return get_session_manager().list(limit=limit, directory=directory)


def delete_session(session_id: str) -> bool:
    """Delete a session."""
    return get_session_manager().delete(session_id)


def fork_session(
    session_id: str,
    from_message_id: str | None = None,
) -> Session | None:
    """Fork a session."""
    return get_session_manager().fork(session_id, from_message_id)



"""
Skill system for PinkyClawd.

Provides specialized instruction sets for common workflows.
Supports automatic discovery of SKILL.md files from:
- .claude/skills/ directories (project and global)
- .opencode/skill/ and .opencode/skills/ directories
"""

from pinkyclawd.skill.loader import (
    discover_and_load_skills,
    discover_skill_md_files,
    load_skill_from_file,
    load_skill_md,
    register_discovered_skills,
)
from pinkyclawd.skill.registry import (
    Skill,
    SkillRegistry,
    get_skill_registry,
)

__all__ = [
    "Skill",
    "SkillRegistry",
    "get_skill_registry",
    "load_skill_from_file",
    "load_skill_md",
    "discover_skill_md_files",
    "discover_and_load_skills",
    "register_discovered_skills",
]



"""
Base agent definition and types.

Agents define different personalities and capabilities for the AI assistant.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Literal


class AgentMode(str, Enum):
    """Mode in which an agent can operate."""

    PRIMARY = "primary"  # Can be selected as main agent
    SUBAGENT = "subagent"  # Only used as a subagent (e.g., Task tool)
    ALL = "all"  # Can be both primary and subagent


@dataclass
class AgentPermissions:
    """Permission configuration for an agent."""

    # File operations
    read: Literal["allow", "ask", "deny"] = "allow"
    write: Literal["allow", "ask", "deny"] = "ask"
    edit: Literal["allow", "ask", "deny"] = "ask"

    # System operations
    bash: Literal["allow", "ask", "deny"] = "ask"
    glob: Literal["allow", "ask", "deny"] = "allow"
    grep: Literal["allow", "ask", "deny"] = "allow"

    # External operations
    webfetch: Literal["allow", "ask", "deny"] = "ask"
    task: Literal["allow", "ask", "deny"] = "allow"

    # Special
    external_directory: Literal["allow", "ask", "deny"] = "ask"

    def to_dict(self) -> dict[str, str]:
        return {
            "read": self.read,
            "write": self.write,
            "edit": self.edit,
            "bash": self.bash,
            "glob": self.glob,
            "grep": self.grep,
            "webfetch": self.webfetch,
            "task": self.task,
            "external_directory": self.external_directory,
        }

    @classmethod
    def from_dict(cls, data: dict[str, str]) -> AgentPermissions:
        return cls(
            read=data.get("read", "allow"),
            write=data.get("write", "ask"),
            edit=data.get("edit", "ask"),
            bash=data.get("bash", "ask"),
            glob=data.get("glob", "allow"),
            grep=data.get("grep", "allow"),
            webfetch=data.get("webfetch", "ask"),
            task=data.get("task", "allow"),
            external_directory=data.get("external_directory", "ask"),
        )

    @classmethod
    def read_only(cls) -> AgentPermissions:
        """Create read-only permissions."""
        return cls(
            read="allow",
            write="deny",
            edit="deny",
            bash="deny",
            glob="allow",
            grep="allow",
            webfetch="deny",
            task="allow",
            external_directory="deny",
        )

    @classmethod
    def full_access(cls) -> AgentPermissions:
        """Create full access permissions (still asks for dangerous ops)."""
        return cls(
            read="allow",
            write="ask",
            edit="ask",
            bash="ask",
            glob="allow",
            grep="allow",
            webfetch="ask",
            task="allow",
            external_directory="ask",
        )


@dataclass
class Agent:
    """
    An agent definition with personality and capabilities.

    Agents customize how the AI assistant behaves for different tasks.
    """

    # Identity
    id: str
    name: str
    description: str

    # Behavior
    system_prompt: str
    mode: AgentMode = AgentMode.PRIMARY
    permissions: AgentPermissions = field(default_factory=AgentPermissions)

    # Model configuration
    model: str | None = None  # Override default model
    temperature: float | None = None
    top_p: float | None = None
    max_steps: int = 100  # Max tool use iterations

    # Display
    color: str = "blue"
    icon: str = ""
    hidden: bool = False

    # Tools
    enabled_tools: list[str] | None = None  # None = all tools
    disabled_tools: list[str] = field(default_factory=list)

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "system_prompt": self.system_prompt,
            "mode": self.mode.value,
            "permissions": self.permissions.to_dict(),
            "model": self.model,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "max_steps": self.max_steps,
            "color": self.color,
            "icon": self.icon,
            "hidden": self.hidden,
            "enabled_tools": self.enabled_tools,
            "disabled_tools": self.disabled_tools,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Agent:
        return cls(
            id=data["id"],
            name=data["name"],
            description=data.get("description", ""),
            system_prompt=data.get("system_prompt", ""),
            mode=AgentMode(data.get("mode", "primary")),
            permissions=AgentPermissions.from_dict(data.get("permissions", {})),
            model=data.get("model"),
            temperature=data.get("temperature"),
            top_p=data.get("top_p"),
            max_steps=data.get("max_steps", 100),
            color=data.get("color", "blue"),
            icon=data.get("icon", ""),
            hidden=data.get("hidden", False),
            enabled_tools=data.get("enabled_tools"),
            disabled_tools=data.get("disabled_tools", []),
        )

    def can_use_tool(self, tool_name: str) -> bool:
        """Check if this agent can use a specific tool."""
        if tool_name in self.disabled_tools:
            return False
        if self.enabled_tools is not None:
            return tool_name in self.enabled_tools
        return True

    def get_permission(self, tool_category: str) -> str:
        """Get permission level for a tool category."""
        return getattr(self.permissions, tool_category, "ask")



"""
Recursive Language Model (RLM) system for unlimited context.

Provides automatic context archival, retrieval, and management
to enable Claude to maintain coherent conversations beyond its
context window limits.

Features:
- Automatic context archival at configurable thresholds
- Pattern-based reference detection for auto-injection
- Semantic search using embeddings (OpenAI, Voyage, local)
- Hybrid search (keyword + semantic + recency) across archived context
- LLM-generated summaries for archived content
- Task-based archival on todo completion
- Session-end archival for context preservation
- RLM-Graph: Knowledge graph storage for structural context retrieval
"""

from pinkyclawd.rlm.context import ContextManager, TokenCounter, get_context_manager
from pinkyclawd.rlm.archive import ContextArchiver, create_archive_block, get_archiver
from pinkyclawd.rlm.search import ContextSearcher, SearchResult, get_searcher
from pinkyclawd.rlm.retrieve import ContextRetriever, RetrievalContext, get_retriever
from pinkyclawd.rlm.embedding import (
    Embedding,
    EmbeddingManager,
    EmbeddingProvider,
    get_embedding_manager,
    generate_embedding,
    generate_embeddings,
    cosine_similarity,
)
from pinkyclawd.rlm.handler import (
    RLMHandler,
    RLMContext,
    get_rlm_handler,
    prepare_messages_with_rlm,
    update_after_response,
    detect_task_completion,
)
from pinkyclawd.rlm.display import (
    configure_display,
    display_retrieval_start,
    display_retrieval_result,
    display_injection,
    display_archival_start,
    display_archival_complete,
    display_context_status,
    register_event_handlers,
)
from pinkyclawd.rlm.auto_inject import (
    AutoInjectResult,
    detects_reference,
    extract_keywords,
    analyze,
    analyze_with_threshold,
    should_auto_inject,
    format_for_injection,
)
from pinkyclawd.rlm.summarize import (
    generate_summary,
    generate_task_summary,
)
from pinkyclawd.rlm.graph import (
    # Types
    NodeType,
    RelationType,
    GraphNode,
    GraphEdge,
    # Components
    EntityExtractor,
    ContentChunker,
    GraphIngester,
    GraphTraverser,
    GraphSearcher,
    GraphStorageAdapter,
    # Accessors
    get_graph_ingester,
    get_graph_traverser,
    get_graph_searcher,
    get_graph_storage,
)
from pinkyclawd.rlm.graph_json import (
    JSONGraphStorage,
    get_json_graph_storage,
    clear_storage_cache,
)

__all__ = [
    # Context management
    "ContextManager",
    "TokenCounter",
    "get_context_manager",
    # Archival
    "ContextArchiver",
    "create_archive_block",
    "get_archiver",
    # Search
    "ContextSearcher",
    "SearchResult",
    "get_searcher",
    # Retrieval
    "ContextRetriever",
    "RetrievalContext",
    "get_retriever",
    # Embeddings (for semantic search)
    "Embedding",
    "EmbeddingManager",
    "EmbeddingProvider",
    "get_embedding_manager",
    "generate_embedding",
    "generate_embeddings",
    "cosine_similarity",
    # Handler (main integration point)
    "RLMHandler",
    "RLMContext",
    "get_rlm_handler",
    "prepare_messages_with_rlm",
    "update_after_response",
    "detect_task_completion",
    # Display
    "configure_display",
    "display_retrieval_start",
    "display_retrieval_result",
    "display_injection",
    "display_archival_start",
    "display_archival_complete",
    "display_context_status",
    "register_event_handlers",
    # Auto-injection
    "AutoInjectResult",
    "detects_reference",
    "extract_keywords",
    "analyze",
    "analyze_with_threshold",
    "should_auto_inject",
    "format_for_injection",
    # Summarization
    "generate_summary",
    "generate_task_summary",
    # Graph (RLM-Graph knowledge graph)
    "NodeType",
    "RelationType",
    "GraphNode",
    "GraphEdge",
    "EntityExtractor",
    "ContentChunker",
    "GraphIngester",
    "GraphTraverser",
    "GraphSearcher",
    "GraphStorageAdapter",
    "get_graph_ingester",
    "get_graph_traverser",
    "get_graph_searcher",
    "get_graph_storage",
    # JSON Storage (file-based graph storage)
    "JSONGraphStorage",
    "get_json_graph_storage",
    "clear_storage_cache",
]



"""
Toast notification system.

Provides non-intrusive notifications that appear temporarily
and dismiss automatically.
"""

from __future__ import annotations

import asyncio
from typing import Any, Literal

from textual.widgets import Static
from textual.containers import Vertical
from textual.app import ComposeResult
from textual.reactive import reactive
from rich.text import Text


ToastLevel = Literal["info", "success", "warning", "error"]


class Toast(Static):
    """A single toast notification."""

    DEFAULT_CSS = """
    Toast {
        width: auto;
        min-width: 30;
        max-width: 60;
        height: auto;
        padding: 1 2;
        margin-bottom: 1;
        background: $surface;
        border: solid $primary;
    }

    Toast.info {
        border-left: thick $accent;
    }

    Toast.success {
        border-left: thick $success;
    }

    Toast.warning {
        border-left: thick $warning;
    }

    Toast.error {
        border-left: thick $error;
    }

    .toast-icon {
        width: 3;
    }

    .toast-message {
        width: 1fr;
    }
    """

    ICONS = {
        "info": "ℹ",
        "success": "✓",
        "warning": "⚠",
        "error": "✗",
    }

    def __init__(
        self,
        message: str,
        level: ToastLevel = "info",
        duration: float = 3.0,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.message = message
        self.level = level
        self.duration = duration
        self.add_class(level)

    def compose(self) -> ComposeResult:
        icon = self.ICONS.get(self.level, "•")
        style = {
            "info": "cyan",
            "success": "green",
            "warning": "yellow",
            "error": "red",
        }.get(self.level, "white")

        text = Text()
        text.append(f"{icon} ", style=style)
        text.append(self.message)
        yield Static(text)


class ToastContainer(Vertical):
    """Container for toast notifications."""

    DEFAULT_CSS = """
    ToastContainer {
        dock: bottom;
        align: right bottom;
        width: auto;
        height: auto;
        max-height: 30%;
        padding: 1;
        layer: notification;
    }
    """

    def show_toast(
        self,
        message: str,
        level: ToastLevel = "info",
        duration: float = 3.0,
    ) -> None:
        """Show a toast notification."""
        toast = Toast(message, level=level, duration=duration)
        self.mount(toast)

        # Schedule removal
        if duration > 0:
            asyncio.create_task(self._remove_after(toast, duration))

    async def _remove_after(self, toast: Toast, duration: float) -> None:
        """Remove toast after duration."""
        await asyncio.sleep(duration)
        if toast.is_mounted:
            toast.remove()

    def clear_all(self) -> None:
        """Clear all toasts."""
        self.remove_children()


# Mixin for apps that want toast support
class ToastMixin:
    """Mixin to add toast notification support to an app."""

    def compose(self) -> ComposeResult:
        """Override to add toast container."""
        yield from super().compose()  # type: ignore
        yield ToastContainer(id="toast-container")

    def show_toast(
        self,
        message: str,
        level: ToastLevel = "info",
        duration: float = 3.0,
    ) -> None:
        """Show a toast notification."""
        try:
            container = self.query_one("#toast-container", ToastContainer)  # type: ignore
            container.show_toast(message, level=level, duration=duration)
        except Exception:
            # Fallback to regular notify
            self.notify(message)  # type: ignore

    def toast_info(self, message: str, duration: float = 3.0) -> None:
        """Show an info toast."""
        self.show_toast(message, "info", duration)

    def toast_success(self, message: str, duration: float = 3.0) -> None:
        """Show a success toast."""
        self.show_toast(message, "success", duration)

    def toast_warning(self, message: str, duration: float = 3.0) -> None:
        """Show a warning toast."""
        self.show_toast(message, "warning", duration)

    def toast_error(self, message: str, duration: float = 3.0) -> None:
        """Show an error toast."""
        self.show_toast(message, "error", duration)



"""
Model Context Protocol (MCP) integration for PinkyClawd.

Provides integration with external MCP servers for extended tool capabilities.
Supports both stdio and HTTP transports with OAuth2 authentication.
"""

from pinkyclawd.mcp.auth import (
    OAuth2Token,
    TokenStore,
    get_token_store,
)
from pinkyclawd.mcp.client import (
    MCPClient,
    MCPServer,
    MCPTool,
    get_mcp_client,
)
from pinkyclawd.mcp.oauth import (
    OAuth2Config,
    OAuth2Flow,
    start_oauth_flow,
    complete_oauth_flow,
)
from pinkyclawd.mcp.transport import (
    HTTPTransport,
    SSETransport,
    StdioTransport,
    Transport,
)

__all__ = [
    # Transport
    "Transport",
    "StdioTransport",
    "HTTPTransport",
    "SSETransport",
    # Auth
    "TokenStore",
    "get_token_store",
    "OAuth2Token",
    # OAuth
    "OAuth2Config",
    "OAuth2Flow",
    "start_oauth_flow",
    "complete_oauth_flow",
    # Client
    "MCPClient",
    "MCPServer",
    "MCPTool",
    "get_mcp_client",
]



"""
Conversation loop with full RLM integration.

This module provides the core conversation loop that integrates:
- Provider communication (Anthropic, OpenAI)
- RLM context management (archival, retrieval, injection)
- Token tracking and automatic compaction
- Task completion detection
"""

from __future__ import annotations

import asyncio
import sys
import uuid
from datetime import datetime
from typing import AsyncIterator

from pinkyclawd.config.settings import get_config
from pinkyclawd.config.storage import (
    Message,
    MessagePart,
    MessageRole,
    PartType,
    Session,
    get_storage,
)
from pinkyclawd.provider.registry import get_provider
from pinkyclawd.rlm import (
    prepare_messages_with_rlm,
    update_after_response,
    detect_task_completion,
    get_rlm_handler,
    display_context_status,
    register_event_handlers,
)


def create_message(
    session_id: str,
    role: MessageRole,
    text: str,
) -> Message:
    """Create a message with text content."""
    msg_id = f"msg_{uuid.uuid4().hex[:12]}"
    return Message(
        id=msg_id,
        session_id=session_id,
        role=role,
        parts=[
            MessagePart(
                id=f"part_{uuid.uuid4().hex[:8]}",
                message_id=msg_id,
                type=PartType.TEXT,
                content={"text": text},
            )
        ],
        created_at=datetime.now(),
    )


def get_message_text(message: Message) -> str:
    """Extract text content from a message."""
    texts = []
    for part in message.parts:
        if part.type == PartType.TEXT:
            texts.append(part.content.get("text", ""))
    return " ".join(texts)


def convert_messages_for_provider(messages: list[Message]) -> list[Message]:
    """
    Convert internal Message objects to provider-compatible format.

    The providers expect Message objects with the same structure,
    so this is mostly a passthrough, but can handle any necessary
    transformations.
    """
    # For now, messages are already in the correct format
    # Future: handle any provider-specific transformations
    return messages


async def run_conversation(
    session_id: str | None = None,
    model: str | None = None,
    initial_prompt: str | None = None,
) -> None:
    """
    Run an interactive conversation loop with RLM integration.

    Args:
        session_id: Optional session ID to continue
        model: Model to use (defaults to config)
        initial_prompt: Optional initial prompt to send
    """
    config = get_config()
    storage = get_storage()
    rlm_handler = get_rlm_handler()

    # Register RLM event handlers for display
    register_event_handlers()

    # Create or load session
    if session_id:
        session = storage.get_session(session_id)
        if not session:
            print(f"Session {session_id} not found, creating new session")
            session_id = None

    if not session_id:
        session_id = f"sess_{uuid.uuid4().hex[:12]}"
        session = Session(
            id=session_id,
            title="New Conversation",
            directory=str(config.working_directory or "."),
        )
        storage.create_session(session)
        print(f"Created session: {session_id}")

    # Load existing messages
    messages = storage.get_messages(session_id)
    print(f"Loaded {len(messages)} existing messages")

    # Get provider
    model = model or config.model
    provider = get_provider(model)
    print(f"Using model: {model}")
    print(f"RLM enabled: {config.rlm.enabled}")
    print("-" * 50)

    # Handle initial prompt if provided
    if initial_prompt:
        await process_turn(
            session_id=session_id,
            user_input=initial_prompt,
            messages=messages,
            provider=provider,
            model=model,
            storage=storage,
            rlm_handler=rlm_handler,
        )

    # Main conversation loop
    while True:
        try:
            # Get user input
            print()
            user_input = input("You: ").strip()

            if not user_input:
                continue

            # Handle special commands
            if user_input.lower() in ("exit", "quit", "/quit", "/exit"):
                # Archive on session end if threshold met
                rlm_handler.archive_on_session_end(session_id, messages)
                print("Goodbye!")
                break

            if user_input == "/context":
                state = rlm_handler.get_context_state(session_id)
                display_context_status(state)
                continue

            if user_input == "/compact":
                if rlm_handler.force_compact(session_id):
                    print("Session compacted successfully")
                else:
                    print("Nothing to compact")
                continue

            if user_input.startswith("/"):
                print(f"Unknown command: {user_input}")
                continue

            # Process the turn
            await process_turn(
                session_id=session_id,
                user_input=user_input,
                messages=messages,
                provider=provider,
                model=model,
                storage=storage,
                rlm_handler=rlm_handler,
            )

        except KeyboardInterrupt:
            print("\nInterrupted")
            break
        except EOFError:
            print("\nEnd of input")
            break


async def process_turn(
    session_id: str,
    user_input: str,
    messages: list[Message],
    provider,
    model: str,
    storage,
    rlm_handler,
) -> None:
    """Process a single conversation turn."""

    # 1. Create and store user message
    user_message = create_message(session_id, MessageRole.USER, user_input)
    messages.append(user_message)
    storage.add_message(user_message)

    # 2. RLM: Prepare messages with context injection
    # This will:
    # - Check for reference patterns that need archived context
    # - Search and retrieve relevant archived blocks
    # - Inject context as a system message
    # - Update token tracking
    print("\n[RLM] Preparing context...")
    augmented_messages, rlm_context = prepare_messages_with_rlm(
        session_id=session_id,
        user_query=user_input,
        messages=messages,
        model=model,
    )

    if rlm_context.has_retrieved_context:
        print(f"[RLM] Injected {len(rlm_context.retrieved_context.blocks)} context blocks")

    # 3. Send to provider and get response
    print("\nAssistant: ", end="", flush=True)

    try:
        # Convert messages to provider format
        provider_messages = convert_messages_for_provider(augmented_messages)

        # Stream the response
        full_response = ""
        async for chunk in provider.stream(
            messages=provider_messages,
            model=model,
        ):
            if chunk.type == "text" and chunk.text:
                print(chunk.text, end="", flush=True)
                full_response += chunk.text
            elif chunk.type == "done":
                break
        print()  # Newline after response

    except Exception as e:
        print(f"\nError: {e}")
        # Remove the user message on error
        messages.pop()
        return

    # 4. Create and store assistant message
    assistant_message = create_message(session_id, MessageRole.ASSISTANT, full_response)
    messages.append(assistant_message)
    storage.add_message(assistant_message)

    # 5. RLM: Update state after response
    # This will:
    # - Update token counts
    # - Trigger archival if threshold (33%) is reached
    context_state = update_after_response(
        session_id=session_id,
        assistant_message=assistant_message,
        model=model,
    )

    # 6. Check for task completion phrases
    # Phrases like "done", "finished", "that's all" trigger archival
    if detect_task_completion(user_input):
        print("[RLM] Task completion detected, archiving context...")
        rlm_handler.check_completion_and_archive(
            message=user_input,
            session_id=session_id,
            messages=messages[-10:],  # Archive recent messages
        )

    # 7. Show context status
    print(f"\n[Context: {context_state.usage_ratio:.1%} ({context_state.total_tokens:,}/{context_state.model_limit:,} tokens)]")


async def run_single_prompt(
    prompt: str,
    model: str | None = None,
    output_format: str = "default",
    max_turns: int = 50,
    no_stream: bool = False,
    output_file: str | None = None,
    working_directory: str | None = None,
) -> dict:
    """
    Run a single prompt (non-interactive mode) with full tool execution.

    Args:
        prompt: The prompt to send
        model: Model to use
        output_format: Output format ("default" or "json")
        max_turns: Maximum number of agent turns (default: 50)
        no_stream: Disable streaming output
        output_file: Optional file to write output to
        working_directory: Working directory for tool execution

    Returns:
        Dict with response and metadata
    """
    import json as json_module
    from pathlib import Path

    from pinkyclawd.tool import get_tool_registry, ToolContext, ToolResult

    config = get_config()
    storage = get_storage()
    rlm_handler = get_rlm_handler()

    # Setup working directory
    work_dir = Path(working_directory) if working_directory else Path.cwd()

    # Create session for this run
    session_id = f"run_{uuid.uuid4().hex[:8]}"
    session = Session(
        id=session_id,
        title=f"Run: {prompt[:50]}...",
        directory=str(work_dir),
    )
    storage.create_session(session)

    # Create user message
    messages = [create_message(session_id, MessageRole.USER, prompt)]

    # Get provider and model
    model = model or config.model
    provider = get_provider(model)

    # Get tool registry
    tool_registry = get_tool_registry()
    tools = tool_registry.get_schemas()

    # Track results
    results = {
        "session_id": session_id,
        "prompt": prompt,
        "model": model,
        "turns": 0,
        "tool_calls": [],
        "response": "",
        "success": True,
        "error": None,
    }

    output_lines = []

    def output(text: str, newline: bool = True) -> None:
        """Write output based on format and streaming settings."""
        if output_format == "default" and not no_stream:
            if newline:
                print(text, flush=True)
            else:
                print(text, end="", flush=True)
        output_lines.append(text)

    # Agent loop
    turn = 0
    while turn < max_turns:
        turn += 1
        results["turns"] = turn

        # Prepare with RLM
        augmented_messages, rlm_context = prepare_messages_with_rlm(
            session_id=session_id,
            user_query=prompt if turn == 1 else "",
            messages=messages,
            model=model,
        )

        # Convert and get response
        provider_messages = convert_messages_for_provider(augmented_messages)

        # Get response with tools
        full_response = ""
        pending_tool_calls = []

        try:
            async for chunk in provider.stream(
                messages=provider_messages,
                model=model,
                tools=tools,
            ):
                if chunk.type == "text" and chunk.text:
                    output(chunk.text, newline=False)
                    full_response += chunk.text
                elif chunk.type == "tool_call_start":
                    pending_tool_calls.append({
                        "id": chunk.tool_call_id,
                        "name": chunk.tool_name,
                        "arguments": "",
                    })
                elif chunk.type == "tool_call_delta":
                    if pending_tool_calls:
                        pending_tool_calls[-1]["arguments"] += chunk.tool_arguments or ""
                elif chunk.type == "tool_call_end":
                    pass  # Tool call complete
                elif chunk.type == "done":
                    break

            if full_response:
                output("")  # Newline after response

        except Exception as e:
            results["success"] = False
            results["error"] = str(e)
            output(f"\nError: {e}")
            break

        # Store assistant message
        assistant_message = create_message(session_id, MessageRole.ASSISTANT, full_response)
        messages.append(assistant_message)
        storage.add_message(assistant_message)
        results["response"] = full_response

        # Update RLM state
        update_after_response(
            session_id=session_id,
            assistant_message=assistant_message,
            model=model,
        )

        # If no tool calls, we're done
        if not pending_tool_calls:
            break

        # Execute tool calls
        tool_results = []
        for tc in pending_tool_calls:
            tool_name = tc["name"]
            try:
                tool_args = json_module.loads(tc["arguments"]) if tc["arguments"] else {}
            except json_module.JSONDecodeError:
                tool_args = {}

            output(f"\n[Tool: {tool_name}]")

            # Create tool context
            ctx = ToolContext(
                session_id=session_id,
                message_id=assistant_message.id,
                working_directory=work_dir,
                user_confirmed=True,  # Auto-confirm in non-interactive mode
            )

            # Execute tool
            try:
                result = await tool_registry.execute(tool_name, ctx, tool_args)
                tool_results.append({
                    "tool_call_id": tc["id"],
                    "success": result.success,
                    "output": result.output,
                    "error": result.error,
                })
                results["tool_calls"].append({
                    "name": tool_name,
                    "arguments": tool_args,
                    "success": result.success,
                })

                # Show truncated output
                output_preview = result.output[:500]
                if len(result.output) > 500:
                    output_preview += "..."
                output(output_preview)

            except Exception as e:
                tool_results.append({
                    "tool_call_id": tc["id"],
                    "success": False,
                    "output": "",
                    "error": str(e),
                })
                output(f"Tool error: {e}")

        # Create tool result message
        tool_parts = []
        for tr in tool_results:
            tool_parts.append(
                MessagePart(
                    id=f"part_{uuid.uuid4().hex[:8]}",
                    message_id=f"tool_{uuid.uuid4().hex[:8]}",
                    type=PartType.TOOL_RESULT,
                    content={
                        "tool_call_id": tr["tool_call_id"],
                        "output": tr["output"],
                        "error": tr["error"],
                    },
                )
            )

        tool_message = Message(
            id=f"msg_tool_{uuid.uuid4().hex[:8]}",
            session_id=session_id,
            role=MessageRole.TOOL,
            parts=tool_parts,
            created_at=datetime.now(),
        )
        messages.append(tool_message)
        storage.add_message(tool_message)

    # Write output file if specified
    if output_file:
        with open(output_file, "w") as f:
            if output_format == "json":
                f.write(json_module.dumps(results, indent=2))
            else:
                f.write("\n".join(output_lines))

    # Print JSON output if requested
    if output_format == "json":
        print(json_module.dumps(results, indent=2))

    return results


def main():
    """CLI entry point for conversation."""
    import argparse

    parser = argparse.ArgumentParser(description="Run PinkyClawd conversation")
    parser.add_argument("--session", "-s", help="Session ID to continue")
    parser.add_argument("--model", "-m", help="Model to use")
    parser.add_argument("--prompt", "-p", help="Initial prompt")
    parser.add_argument("--run", "-r", help="Run single prompt (non-interactive)")

    args = parser.parse_args()

    if args.run:
        # Non-interactive mode
        asyncio.run(run_single_prompt(args.run, args.model))
    else:
        # Interactive mode
        asyncio.run(run_conversation(args.session, args.model, args.prompt))


if __name__ == "__main__":
    main()



"""
HTTP Server for PinkyClawd.

Provides a REST API and WebSocket interface for remote access.
"""

from pinkyclawd.server.app import create_app, run_server

__all__ = [
    "create_app",
    "run_server",
]



"""
SQLite storage for sessions, messages, and RLM context.

Provides persistent storage matching pinkyclawd's data model.
"""

from __future__ import annotations

import json
import sqlite3
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Generator, Literal
from enum import Enum

from pinkyclawd.config.paths import get_database_path


class MessageRole(str, Enum):
    """Message roles."""

    USER = "user"
    ASSISTANT = "assistant"
    SYSTEM = "system"
    TOOL = "tool"


class PartType(str, Enum):
    """Message part types."""

    TEXT = "text"
    TOOL_USE = "tool_use"
    TOOL_RESULT = "tool_result"
    IMAGE = "image"
    FILE = "file"


@dataclass
class MessagePart:
    """A part of a message (text, tool call, etc.)."""

    id: str
    message_id: str
    type: PartType
    content: dict[str, Any]
    created_at: datetime = field(default_factory=datetime.now)

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "message_id": self.message_id,
            "type": self.type.value,
            "content": self.content,
            "created_at": self.created_at.isoformat(),
        }

    @classmethod
    def from_dict(cls, data: dict) -> MessagePart:
        return cls(
            id=data["id"],
            message_id=data["message_id"],
            type=PartType(data["type"]),
            content=data["content"],
            created_at=datetime.fromisoformat(data["created_at"]),
        )


@dataclass
class Message:
    """A message in a session."""

    id: str
    session_id: str
    role: MessageRole
    parts: list[MessagePart] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.now)
    metadata: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "session_id": self.session_id,
            "role": self.role.value,
            "parts": [p.to_dict() for p in self.parts],
            "created_at": self.created_at.isoformat(),
            "metadata": self.metadata,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Message:
        return cls(
            id=data["id"],
            session_id=data["session_id"],
            role=MessageRole(data["role"]),
            parts=[MessagePart.from_dict(p) for p in data.get("parts", [])],
            created_at=datetime.fromisoformat(data["created_at"]),
            metadata=data.get("metadata", {}),
        )


class SessionStatus(str, Enum):
    """Session status states."""

    IDLE = "idle"
    ACTIVE = "active"
    ABORTED = "aborted"


@dataclass
class Session:
    """A conversation session."""

    id: str
    title: str = "New Session"
    directory: str = ""
    parent_id: str | None = None
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    archived_at: datetime | None = None
    share_token: str | None = None
    status: SessionStatus = SessionStatus.IDLE
    metadata: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "title": self.title,
            "directory": self.directory,
            "parent_id": self.parent_id,
            "created_at": self.created_at.isoformat(),
            "updated_at": self.updated_at.isoformat(),
            "archived_at": self.archived_at.isoformat() if self.archived_at else None,
            "share_token": self.share_token,
            "status": self.status.value,
            "metadata": self.metadata,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Session:
        return cls(
            id=data["id"],
            title=data.get("title", "New Session"),
            directory=data.get("directory", ""),
            parent_id=data.get("parent_id"),
            created_at=datetime.fromisoformat(data["created_at"]),
            updated_at=datetime.fromisoformat(data["updated_at"]),
            archived_at=datetime.fromisoformat(data["archived_at"])
            if data.get("archived_at")
            else None,
            share_token=data.get("share_token"),
            status=SessionStatus(data.get("status", "idle")),
            metadata=data.get("metadata", {}),
        )


@dataclass
class ContextBlock:
    """An archived context block for RLM."""

    id: str
    session_id: str
    task_id: str | None = None
    task_description: str = ""
    summary: str = ""
    content: str = ""
    tokens: int = 0
    created_at: datetime = field(default_factory=datetime.now)

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "session_id": self.session_id,
            "task_id": self.task_id,
            "task_description": self.task_description,
            "summary": self.summary,
            "content": self.content,
            "tokens": self.tokens,
            "created_at": self.created_at.isoformat(),
        }

    @classmethod
    def from_dict(cls, data: dict) -> ContextBlock:
        return cls(
            id=data["id"],
            session_id=data["session_id"],
            task_id=data.get("task_id"),
            task_description=data.get("task_description", ""),
            summary=data.get("summary", ""),
            content=data.get("content", ""),
            tokens=data.get("tokens", 0),
            created_at=datetime.fromisoformat(data["created_at"]),
        )


class Storage:
    """SQLite-based storage for PinkyClawd data."""

    def __init__(self, db_path: Path | None = None) -> None:
        self.db_path = db_path or get_database_path()
        self._init_db()

    def _init_db(self) -> None:
        """Initialize database schema."""
        with self._connect() as conn:
            conn.executescript("""
                CREATE TABLE IF NOT EXISTS sessions (
                    id TEXT PRIMARY KEY,
                    title TEXT NOT NULL DEFAULT 'New Session',
                    directory TEXT DEFAULT '',
                    parent_id TEXT,
                    created_at TEXT NOT NULL,
                    updated_at TEXT NOT NULL,
                    archived_at TEXT,
                    share_token TEXT,
                    status TEXT DEFAULT 'idle',
                    metadata TEXT DEFAULT '{}',
                    FOREIGN KEY (parent_id) REFERENCES sessions(id)
                );
                
                CREATE TABLE IF NOT EXISTS messages (
                    id TEXT PRIMARY KEY,
                    session_id TEXT NOT NULL,
                    role TEXT NOT NULL,
                    created_at TEXT NOT NULL,
                    metadata TEXT DEFAULT '{}',
                    FOREIGN KEY (session_id) REFERENCES sessions(id) ON DELETE CASCADE
                );
                
                CREATE TABLE IF NOT EXISTS message_parts (
                    id TEXT PRIMARY KEY,
                    message_id TEXT NOT NULL,
                    type TEXT NOT NULL,
                    content TEXT NOT NULL,
                    created_at TEXT NOT NULL,
                    FOREIGN KEY (message_id) REFERENCES messages(id) ON DELETE CASCADE
                );
                
                CREATE TABLE IF NOT EXISTS context_blocks (
                    id TEXT PRIMARY KEY,
                    session_id TEXT NOT NULL,
                    task_id TEXT,
                    task_description TEXT DEFAULT '',
                    summary TEXT DEFAULT '',
                    content TEXT NOT NULL,
                    tokens INTEGER DEFAULT 0,
                    created_at TEXT NOT NULL,
                    FOREIGN KEY (session_id) REFERENCES sessions(id) ON DELETE CASCADE
                );
                
                CREATE TABLE IF NOT EXISTS todos (
                    id TEXT PRIMARY KEY,
                    session_id TEXT NOT NULL,
                    content TEXT NOT NULL,
                    status TEXT DEFAULT 'pending',
                    priority TEXT DEFAULT 'medium',
                    created_at TEXT NOT NULL,
                    updated_at TEXT NOT NULL,
                    FOREIGN KEY (session_id) REFERENCES sessions(id) ON DELETE CASCADE
                );
                
                CREATE INDEX IF NOT EXISTS idx_messages_session ON messages(session_id);
                CREATE INDEX IF NOT EXISTS idx_parts_message ON message_parts(message_id);
                CREATE INDEX IF NOT EXISTS idx_context_session ON context_blocks(session_id);
                CREATE INDEX IF NOT EXISTS idx_todos_session ON todos(session_id);

                CREATE TABLE IF NOT EXISTS embeddings (
                    block_id TEXT PRIMARY KEY,
                    vector TEXT NOT NULL,
                    model TEXT NOT NULL,
                    dimensions INTEGER NOT NULL,
                    content_hash TEXT NOT NULL,
                    created_at TEXT NOT NULL,
                    FOREIGN KEY (block_id) REFERENCES context_blocks(id) ON DELETE CASCADE
                );

                CREATE INDEX IF NOT EXISTS idx_embeddings_block ON embeddings(block_id);

                -- RLM-Graph tables for knowledge graph storage
                CREATE TABLE IF NOT EXISTS graph_nodes (
                    id TEXT PRIMARY KEY,
                    session_id TEXT NOT NULL,
                    type TEXT NOT NULL,  -- 'document', 'section', 'chunk', 'entity'
                    content TEXT NOT NULL,
                    metadata TEXT DEFAULT '{}',
                    embedding BLOB,
                    created_at TEXT NOT NULL,
                    FOREIGN KEY (session_id) REFERENCES sessions(id) ON DELETE CASCADE
                );

                CREATE TABLE IF NOT EXISTS graph_edges (
                    source_id TEXT NOT NULL,
                    target_id TEXT NOT NULL,
                    relationship TEXT NOT NULL,  -- 'has_section', 'has_chunk', 'mentions', 'related_to'
                    weight REAL DEFAULT 1.0,
                    metadata TEXT DEFAULT '{}',
                    created_at TEXT NOT NULL,
                    PRIMARY KEY (source_id, target_id, relationship),
                    FOREIGN KEY (source_id) REFERENCES graph_nodes(id) ON DELETE CASCADE,
                    FOREIGN KEY (target_id) REFERENCES graph_nodes(id) ON DELETE CASCADE
                );

                CREATE INDEX IF NOT EXISTS idx_graph_nodes_session ON graph_nodes(session_id);
                CREATE INDEX IF NOT EXISTS idx_graph_nodes_type ON graph_nodes(type);
                CREATE INDEX IF NOT EXISTS idx_graph_edges_source ON graph_edges(source_id);
                CREATE INDEX IF NOT EXISTS idx_graph_edges_target ON graph_edges(target_id);
                CREATE INDEX IF NOT EXISTS idx_graph_edges_rel ON graph_edges(relationship);
            """)

    @contextmanager
    def _connect(self) -> Generator[sqlite3.Connection, None, None]:
        """Get a database connection."""
        conn = sqlite3.connect(self.db_path)
        conn.row_factory = sqlite3.Row
        try:
            yield conn
            conn.commit()
        except Exception:
            conn.rollback()
            raise
        finally:
            conn.close()

    # Session operations

    def create_session(self, session: Session) -> Session:
        """Create a new session."""
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO sessions (id, title, directory, parent_id, created_at, updated_at, share_token, status, metadata)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    session.id,
                    session.title,
                    session.directory,
                    session.parent_id,
                    session.created_at.isoformat(),
                    session.updated_at.isoformat(),
                    session.share_token,
                    session.status.value,
                    json.dumps(session.metadata),
                ),
            )
        return session

    def get_session(self, session_id: str) -> Session | None:
        """Get a session by ID."""
        with self._connect() as conn:
            row = conn.execute(
                "SELECT * FROM sessions WHERE id = ?",
                (session_id,),
            ).fetchone()
            if row:
                return self._row_to_session(row)
        return None

    def _row_to_session(self, row: sqlite3.Row) -> Session:
        """Convert a database row to a Session object."""
        return Session(
            id=row["id"],
            title=row["title"],
            directory=row["directory"],
            parent_id=row["parent_id"],
            created_at=datetime.fromisoformat(row["created_at"]),
            updated_at=datetime.fromisoformat(row["updated_at"]),
            archived_at=datetime.fromisoformat(row["archived_at"])
            if row["archived_at"]
            else None,
            share_token=row["share_token"] if "share_token" in row.keys() else None,
            status=SessionStatus(row["status"]) if "status" in row.keys() and row["status"] else SessionStatus.IDLE,
            metadata=json.loads(row["metadata"]),
        )

    def list_sessions(
        self,
        limit: int = 50,
        offset: int = 0,
        directory: str | None = None,
    ) -> list[Session]:
        """List sessions with optional filtering."""
        with self._connect() as conn:
            query = "SELECT * FROM sessions"
            params: list[Any] = []

            if directory:
                query += " WHERE directory = ?"
                params.append(directory)

            query += " ORDER BY updated_at DESC LIMIT ? OFFSET ?"
            params.extend([limit, offset])

            rows = conn.execute(query, params).fetchall()
            return [self._row_to_session(row) for row in rows]

    def update_session(self, session: Session) -> None:
        """Update a session."""
        session.updated_at = datetime.now()
        with self._connect() as conn:
            conn.execute(
                """
                UPDATE sessions
                SET title = ?, directory = ?, parent_id = ?, updated_at = ?,
                    archived_at = ?, share_token = ?, status = ?, metadata = ?
                WHERE id = ?
                """,
                (
                    session.title,
                    session.directory,
                    session.parent_id,
                    session.updated_at.isoformat(),
                    session.archived_at.isoformat() if session.archived_at else None,
                    session.share_token,
                    session.status.value,
                    json.dumps(session.metadata),
                    session.id,
                ),
            )

    def delete_session(self, session_id: str) -> None:
        """Delete a session and all its data."""
        with self._connect() as conn:
            conn.execute("DELETE FROM sessions WHERE id = ?", (session_id,))

    # Message operations

    def add_message(self, message: Message) -> Message:
        """Add a message to a session."""
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO messages (id, session_id, role, created_at, metadata)
                VALUES (?, ?, ?, ?, ?)
                """,
                (
                    message.id,
                    message.session_id,
                    message.role.value,
                    message.created_at.isoformat(),
                    json.dumps(message.metadata),
                ),
            )

            for part in message.parts:
                conn.execute(
                    """
                    INSERT INTO message_parts (id, message_id, type, content, created_at)
                    VALUES (?, ?, ?, ?, ?)
                    """,
                    (
                        part.id,
                        part.message_id,
                        part.type.value,
                        json.dumps(part.content),
                        part.created_at.isoformat(),
                    ),
                )
        return message

    def get_messages(self, session_id: str) -> list[Message]:
        """Get all messages for a session."""
        with self._connect() as conn:
            msg_rows = conn.execute(
                "SELECT * FROM messages WHERE session_id = ? ORDER BY created_at",
                (session_id,),
            ).fetchall()

            messages = []
            for msg_row in msg_rows:
                part_rows = conn.execute(
                    "SELECT * FROM message_parts WHERE message_id = ? ORDER BY created_at",
                    (msg_row["id"],),
                ).fetchall()

                parts = [
                    MessagePart(
                        id=p["id"],
                        message_id=p["message_id"],
                        type=PartType(p["type"]),
                        content=json.loads(p["content"]),
                        created_at=datetime.fromisoformat(p["created_at"]),
                    )
                    for p in part_rows
                ]

                messages.append(
                    Message(
                        id=msg_row["id"],
                        session_id=msg_row["session_id"],
                        role=MessageRole(msg_row["role"]),
                        parts=parts,
                        created_at=datetime.fromisoformat(msg_row["created_at"]),
                        metadata=json.loads(msg_row["metadata"]),
                    )
                )

            return messages

    def add_message_part(self, part: MessagePart) -> None:
        """Add a part to an existing message."""
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO message_parts (id, message_id, type, content, created_at)
                VALUES (?, ?, ?, ?, ?)
                """,
                (
                    part.id,
                    part.message_id,
                    part.type.value,
                    json.dumps(part.content),
                    part.created_at.isoformat(),
                ),
            )

    def get_message(self, message_id: str) -> Message | None:
        """Get a message by ID."""
        with self._connect() as conn:
            msg_row = conn.execute(
                "SELECT * FROM messages WHERE id = ?",
                (message_id,),
            ).fetchone()

            if not msg_row:
                return None

            part_rows = conn.execute(
                "SELECT * FROM message_parts WHERE message_id = ? ORDER BY created_at",
                (message_id,),
            ).fetchall()

            parts = [
                MessagePart(
                    id=p["id"],
                    message_id=p["message_id"],
                    type=PartType(p["type"]),
                    content=json.loads(p["content"]),
                    created_at=datetime.fromisoformat(p["created_at"]),
                )
                for p in part_rows
            ]

            return Message(
                id=msg_row["id"],
                session_id=msg_row["session_id"],
                role=MessageRole(msg_row["role"]),
                parts=parts,
                created_at=datetime.fromisoformat(msg_row["created_at"]),
                metadata=json.loads(msg_row["metadata"]),
            )

    def get_message_part(self, part_id: str) -> MessagePart | None:
        """Get a message part by ID."""
        with self._connect() as conn:
            row = conn.execute(
                "SELECT * FROM message_parts WHERE id = ?",
                (part_id,),
            ).fetchone()

            if row:
                return MessagePart(
                    id=row["id"],
                    message_id=row["message_id"],
                    type=PartType(row["type"]),
                    content=json.loads(row["content"]),
                    created_at=datetime.fromisoformat(row["created_at"]),
                )
        return None

    def update_message_part(self, part: MessagePart) -> None:
        """Update a message part."""
        with self._connect() as conn:
            conn.execute(
                """
                UPDATE message_parts SET type = ?, content = ? WHERE id = ?
                """,
                (
                    part.type.value,
                    json.dumps(part.content),
                    part.id,
                ),
            )

    def delete_message_part(self, part_id: str) -> bool:
        """Delete a message part by ID."""
        with self._connect() as conn:
            cursor = conn.execute(
                "DELETE FROM message_parts WHERE id = ?",
                (part_id,),
            )
            return cursor.rowcount > 0

    def delete_message(self, message_id: str) -> bool:
        """Delete a message and its parts."""
        with self._connect() as conn:
            cursor = conn.execute(
                "DELETE FROM messages WHERE id = ?",
                (message_id,),
            )
            return cursor.rowcount > 0

    def get_sessions_by_status(self, status: SessionStatus) -> list[Session]:
        """Get all sessions with a specific status."""
        with self._connect() as conn:
            rows = conn.execute(
                "SELECT * FROM sessions WHERE status = ? ORDER BY updated_at DESC",
                (status.value,),
            ).fetchall()
            return [self._row_to_session(row) for row in rows]

    def get_session_by_share_token(self, share_token: str) -> Session | None:
        """Get a session by its share token."""
        with self._connect() as conn:
            row = conn.execute(
                "SELECT * FROM sessions WHERE share_token = ?",
                (share_token,),
            ).fetchone()
            if row:
                return self._row_to_session(row)
        return None

    def get_child_sessions(self, parent_id: str) -> list[Session]:
        """Get all child sessions of a parent."""
        with self._connect() as conn:
            rows = conn.execute(
                "SELECT * FROM sessions WHERE parent_id = ? ORDER BY created_at DESC",
                (parent_id,),
            ).fetchall()
            return [self._row_to_session(row) for row in rows]

    # Context block operations (for RLM)

    def add_context_block(self, block: ContextBlock) -> ContextBlock:
        """Add a context block."""
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO context_blocks 
                (id, session_id, task_id, task_description, summary, content, tokens, created_at)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    block.id,
                    block.session_id,
                    block.task_id,
                    block.task_description,
                    block.summary,
                    block.content,
                    block.tokens,
                    block.created_at.isoformat(),
                ),
            )
        return block

    def get_context_blocks(
        self,
        session_id: str | None = None,
        limit: int = 100,
    ) -> list[ContextBlock]:
        """Get context blocks, optionally filtered by session."""
        with self._connect() as conn:
            if session_id:
                rows = conn.execute(
                    "SELECT * FROM context_blocks WHERE session_id = ? ORDER BY created_at DESC LIMIT ?",
                    (session_id, limit),
                ).fetchall()
            else:
                rows = conn.execute(
                    "SELECT * FROM context_blocks ORDER BY created_at DESC LIMIT ?",
                    (limit,),
                ).fetchall()

            return [
                ContextBlock(
                    id=row["id"],
                    session_id=row["session_id"],
                    task_id=row["task_id"],
                    task_description=row["task_description"],
                    summary=row["summary"],
                    content=row["content"],
                    tokens=row["tokens"],
                    created_at=datetime.fromisoformat(row["created_at"]),
                )
                for row in rows
            ]

    def search_context_blocks(self, query: str, limit: int = 10) -> list[ContextBlock]:
        """Search context blocks by content."""
        with self._connect() as conn:
            rows = conn.execute(
                """
                SELECT * FROM context_blocks
                WHERE content LIKE ? OR summary LIKE ? OR task_description LIKE ?
                ORDER BY created_at DESC LIMIT ?
                """,
                (f"%{query}%", f"%{query}%", f"%{query}%", limit),
            ).fetchall()

            return [
                ContextBlock(
                    id=row["id"],
                    session_id=row["session_id"],
                    task_id=row["task_id"],
                    task_description=row["task_description"],
                    summary=row["summary"],
                    content=row["content"],
                    tokens=row["tokens"],
                    created_at=datetime.fromisoformat(row["created_at"]),
                )
                for row in rows
            ]

    # Embedding operations (for semantic search)

    def save_embedding(
        self,
        block_id: str,
        vector: list[float],
        model: str,
        dimensions: int,
        content_hash: str,
    ) -> None:
        """Save an embedding for a context block."""
        with self._connect() as conn:
            conn.execute(
                """
                INSERT OR REPLACE INTO embeddings
                (block_id, vector, model, dimensions, content_hash, created_at)
                VALUES (?, ?, ?, ?, ?, ?)
                """,
                (
                    block_id,
                    json.dumps(vector),
                    model,
                    dimensions,
                    content_hash,
                    datetime.now().isoformat(),
                ),
            )

    def get_embedding(self, block_id: str) -> tuple[list[float], str, str] | None:
        """
        Get embedding for a context block.

        Returns:
            Tuple of (vector, model, content_hash) or None if not found
        """
        with self._connect() as conn:
            row = conn.execute(
                "SELECT vector, model, content_hash FROM embeddings WHERE block_id = ?",
                (block_id,),
            ).fetchone()

            if row:
                return (
                    json.loads(row["vector"]),
                    row["model"],
                    row["content_hash"],
                )
        return None

    def get_all_embeddings(
        self,
        session_id: str | None = None,
    ) -> list[tuple[str, list[float], str]]:
        """
        Get all embeddings, optionally filtered by session.

        Returns:
            List of (block_id, vector, content_hash) tuples
        """
        with self._connect() as conn:
            if session_id:
                rows = conn.execute(
                    """
                    SELECT e.block_id, e.vector, e.content_hash
                    FROM embeddings e
                    JOIN context_blocks cb ON e.block_id = cb.id
                    WHERE cb.session_id = ?
                    """,
                    (session_id,),
                ).fetchall()
            else:
                rows = conn.execute(
                    "SELECT block_id, vector, content_hash FROM embeddings"
                ).fetchall()

            return [
                (row["block_id"], json.loads(row["vector"]), row["content_hash"])
                for row in rows
            ]

    def delete_embedding(self, block_id: str) -> None:
        """Delete embedding for a context block."""
        with self._connect() as conn:
            conn.execute("DELETE FROM embeddings WHERE block_id = ?", (block_id,))

    def has_embedding(self, block_id: str, content_hash: str) -> bool:
        """Check if a valid embedding exists for a block."""
        with self._connect() as conn:
            row = conn.execute(
                "SELECT content_hash FROM embeddings WHERE block_id = ?",
                (block_id,),
            ).fetchone()
            return row is not None and row["content_hash"] == content_hash

    # Graph operations (for RLM-Graph)

    def add_graph_node(
        self,
        node_id: str,
        session_id: str,
        node_type: str,
        content: str,
        metadata: dict[str, Any] | None = None,
        embedding: list[float] | None = None,
    ) -> None:
        """Add a node to the knowledge graph."""
        with self._connect() as conn:
            conn.execute(
                """
                INSERT OR REPLACE INTO graph_nodes
                (id, session_id, type, content, metadata, embedding, created_at)
                VALUES (?, ?, ?, ?, ?, ?, ?)
                """,
                (
                    node_id,
                    session_id,
                    node_type,
                    content,
                    json.dumps(metadata or {}),
                    json.dumps(embedding) if embedding else None,
                    datetime.now().isoformat(),
                ),
            )

    def get_graph_node(self, node_id: str) -> dict[str, Any] | None:
        """Get a node from the knowledge graph."""
        with self._connect() as conn:
            row = conn.execute(
                "SELECT * FROM graph_nodes WHERE id = ?",
                (node_id,),
            ).fetchone()

            if row:
                return {
                    "id": row["id"],
                    "session_id": row["session_id"],
                    "type": row["type"],
                    "content": row["content"],
                    "metadata": json.loads(row["metadata"]),
                    "embedding": json.loads(row["embedding"]) if row["embedding"] else None,
                    "created_at": row["created_at"],
                }
        return None

    def get_graph_nodes(
        self,
        session_id: str | None = None,
        node_type: str | None = None,
        limit: int = 100,
    ) -> list[dict[str, Any]]:
        """Get nodes from the knowledge graph with optional filters."""
        with self._connect() as conn:
            query = "SELECT * FROM graph_nodes WHERE 1=1"
            params: list[Any] = []

            if session_id:
                query += " AND session_id = ?"
                params.append(session_id)

            if node_type:
                query += " AND type = ?"
                params.append(node_type)

            query += " ORDER BY created_at DESC LIMIT ?"
            params.append(limit)

            rows = conn.execute(query, params).fetchall()

            return [
                {
                    "id": row["id"],
                    "session_id": row["session_id"],
                    "type": row["type"],
                    "content": row["content"],
                    "metadata": json.loads(row["metadata"]),
                    "embedding": json.loads(row["embedding"]) if row["embedding"] else None,
                    "created_at": row["created_at"],
                }
                for row in rows
            ]

    def search_graph_nodes(
        self,
        query: str,
        session_id: str | None = None,
        node_type: str | None = None,
        limit: int = 20,
    ) -> list[dict[str, Any]]:
        """Search nodes by content."""
        with self._connect() as conn:
            sql = "SELECT * FROM graph_nodes WHERE content LIKE ?"
            params: list[Any] = [f"%{query}%"]

            if session_id:
                sql += " AND session_id = ?"
                params.append(session_id)

            if node_type:
                sql += " AND type = ?"
                params.append(node_type)

            sql += " ORDER BY created_at DESC LIMIT ?"
            params.append(limit)

            rows = conn.execute(sql, params).fetchall()

            return [
                {
                    "id": row["id"],
                    "session_id": row["session_id"],
                    "type": row["type"],
                    "content": row["content"],
                    "metadata": json.loads(row["metadata"]),
                    "embedding": json.loads(row["embedding"]) if row["embedding"] else None,
                    "created_at": row["created_at"],
                }
                for row in rows
            ]

    def delete_graph_node(self, node_id: str) -> bool:
        """Delete a node and its edges."""
        with self._connect() as conn:
            cursor = conn.execute("DELETE FROM graph_nodes WHERE id = ?", (node_id,))
            return cursor.rowcount > 0

    def add_graph_edge(
        self,
        source_id: str,
        target_id: str,
        relationship: str,
        weight: float = 1.0,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Add an edge to the knowledge graph."""
        with self._connect() as conn:
            conn.execute(
                """
                INSERT OR REPLACE INTO graph_edges
                (source_id, target_id, relationship, weight, metadata, created_at)
                VALUES (?, ?, ?, ?, ?, ?)
                """,
                (
                    source_id,
                    target_id,
                    relationship,
                    weight,
                    json.dumps(metadata or {}),
                    datetime.now().isoformat(),
                ),
            )

    def get_graph_edges(
        self,
        node_id: str,
        direction: Literal["outgoing", "incoming", "both"] = "both",
        relationship: str | None = None,
    ) -> list[dict[str, Any]]:
        """Get edges connected to a node."""
        with self._connect() as conn:
            edges = []

            if direction in ("outgoing", "both"):
                query = "SELECT * FROM graph_edges WHERE source_id = ?"
                params: list[Any] = [node_id]
                if relationship:
                    query += " AND relationship = ?"
                    params.append(relationship)

                rows = conn.execute(query, params).fetchall()
                edges.extend([
                    {
                        "source_id": row["source_id"],
                        "target_id": row["target_id"],
                        "relationship": row["relationship"],
                        "weight": row["weight"],
                        "metadata": json.loads(row["metadata"]),
                        "direction": "outgoing",
                    }
                    for row in rows
                ])

            if direction in ("incoming", "both"):
                query = "SELECT * FROM graph_edges WHERE target_id = ?"
                params = [node_id]
                if relationship:
                    query += " AND relationship = ?"
                    params.append(relationship)

                rows = conn.execute(query, params).fetchall()
                edges.extend([
                    {
                        "source_id": row["source_id"],
                        "target_id": row["target_id"],
                        "relationship": row["relationship"],
                        "weight": row["weight"],
                        "metadata": json.loads(row["metadata"]),
                        "direction": "incoming",
                    }
                    for row in rows
                ])

            return edges

    def get_neighbors(
        self,
        node_id: str,
        relationship: str | None = None,
        direction: Literal["outgoing", "incoming", "both"] = "outgoing",
        node_type: str | None = None,
    ) -> list[dict[str, Any]]:
        """Get neighboring nodes connected by edges."""
        edges = self.get_graph_edges(node_id, direction, relationship)
        neighbor_ids = set()

        for edge in edges:
            if edge["direction"] == "outgoing":
                neighbor_ids.add(edge["target_id"])
            else:
                neighbor_ids.add(edge["source_id"])

        if not neighbor_ids:
            return []

        with self._connect() as conn:
            placeholders = ",".join("?" * len(neighbor_ids))
            query = f"SELECT * FROM graph_nodes WHERE id IN ({placeholders})"
            params: list[Any] = list(neighbor_ids)

            if node_type:
                query += " AND type = ?"
                params.append(node_type)

            rows = conn.execute(query, params).fetchall()

            return [
                {
                    "id": row["id"],
                    "session_id": row["session_id"],
                    "type": row["type"],
                    "content": row["content"],
                    "metadata": json.loads(row["metadata"]),
                    "embedding": json.loads(row["embedding"]) if row["embedding"] else None,
                    "created_at": row["created_at"],
                }
                for row in rows
            ]

    def delete_graph_edge(
        self,
        source_id: str,
        target_id: str,
        relationship: str,
    ) -> bool:
        """Delete an edge from the knowledge graph."""
        with self._connect() as conn:
            cursor = conn.execute(
                "DELETE FROM graph_edges WHERE source_id = ? AND target_id = ? AND relationship = ?",
                (source_id, target_id, relationship),
            )
            return cursor.rowcount > 0

    def clear_session_graph(self, session_id: str) -> None:
        """Clear all graph data for a session."""
        with self._connect() as conn:
            # Get node IDs first
            node_ids = conn.execute(
                "SELECT id FROM graph_nodes WHERE session_id = ?",
                (session_id,),
            ).fetchall()

            if node_ids:
                ids = [row["id"] for row in node_ids]
                placeholders = ",".join("?" * len(ids))

                # Delete edges connected to these nodes
                conn.execute(
                    f"DELETE FROM graph_edges WHERE source_id IN ({placeholders}) OR target_id IN ({placeholders})",
                    ids + ids,
                )

            # Delete nodes
            conn.execute("DELETE FROM graph_nodes WHERE session_id = ?", (session_id,))


# Global storage instance
_storage: Storage | None = None


def get_storage() -> Storage:
    """Get the global storage instance."""
    global _storage
    if _storage is None:
        _storage = Storage()
    return _storage
